#  Copyright (c) 2024 zfit

from __future__ import annotations

from typing import TYPE_CHECKING

from tensorflow.python.util.deprecation import deprecated_args

from ..util.plotter import PDFPlotter

if TYPE_CHECKING:
    import zfit

import operator
from collections.abc import Callable, Iterable
from contextlib import suppress
from functools import reduce

import numpy as np
import tensorflow as tf
import tensorflow_probability as tfp
from uhi.typing.plottable import PlottableHistogram

import zfit
import zfit.z.numpy as znp
from zfit import z
from zfit._data.binneddatav1 import BinnedData, BinnedSamplerData, move_axis_obs

from ..util import ztyping
from ..util.cache import GraphCachable
from ..util.container import convert_to_container
from ..util.deprecation import deprecated, deprecated_norm_range
from ..util.exception import (
    AlreadyExtendedPDFError,
    BasePDFSubclassingError,
    BreakingAPIChangeError,
    MultipleLimitsNotImplemented,
    NormNotImplemented,
    NotExtendedPDFError,
    SpaceIncompatibleError,
    SpecificFunctionNotImplemented,
    WorkInProgressError,
)
from .baseobject import BaseNumeric, extract_filter_params
from .binning import unbinned_to_binindex
from .data import Data
from .dimension import BaseDimensional
from .interfaces import (
    ZfitBinnedData,
    ZfitBinnedPDF,
    ZfitBinning,
    ZfitParameter,
    ZfitPDF,
    ZfitSpace,
    ZfitUnbinnedData,
)
from .parameter import convert_to_parameter
from .space import convert_to_space, supports
from .tensorlike import OverloadableMixinValues

_BaseModel_USER_IMPL_METHODS_TO_CHECK = {}


def _BinnedPDF_register_check_support(has_support: bool):
    """Marks a method that the subclass either *has* to or *can't* use the ``@supports`` decorator.

    Args:
        has_support (bool): If True, flags that it **requires** the ``@supports`` decorator. If False,
            flags that the ``@supports`` decorator is **not allowed**.
    """
    if not isinstance(has_support, bool):
        msg = "Has to be boolean."
        raise TypeError(msg)

    def register(func):
        """Register a method to be checked to (if True) *has* ``support`` or (if False) has *no* ``support``.

        Args:
            func (function):
        Returns:
            function:
        """
        name = func.__name__
        _BaseModel_USER_IMPL_METHODS_TO_CHECK[name] = has_support
        func.__wrapped__ = _BinnedPDF_register_check_support
        return func

    return register


class BaseBinnedPDF(
    BaseNumeric,
    GraphCachable,
    BaseDimensional,
    OverloadableMixinValues,
    ZfitBinnedPDF,
):
    def __init__(
        self,
        obs: ztyping.ObsTypeInput,
        extended: ztyping.ExtendedInputType = None,
        norm: ztyping.NormInputType = None,
        name: str | None = None,
        label: str | None = None,
        **kwargs,
    ):
        self._label = label
        self.plot = None
        super().__init__(dtype=znp.float64, name=name, **kwargs)

        self._space = self._check_convert_obs_init(obs)
        self._yield = None
        self._norm = self._check_convert_norm_init(norm)
        if extended is None:
            extended = False
        if extended is not False:
            self._set_yield(extended)

        if self.plot is None:  # todo: have a binned plotter
            self.plot = PDFPlotter(self)

    @property
    def label(self):
        return self._label

    def _check_convert_obs_init(self, obs):
        if not isinstance(obs, ZfitSpace) or not obs.is_binned:
            msg = f"`obs` have to be a Space with binning, not {obs}."
            raise ValueError(msg)
        return obs

    def _check_convert_norm_init(self, norm):
        if norm is not None and (not isinstance(norm, ZfitSpace) or not norm.has_limits):
            msg = f"`norm` has to be None or a Space with limits, not {norm}."
            raise ValueError(msg)
        return norm

    @classmethod
    def _subclass_check_support(cls, methods_to_check, wrapper_not_overwritten):
        for method_name, has_support in methods_to_check.items():
            method = getattr(cls, method_name)
            if hasattr(method, "__wrapped__") and method.__wrapped__ == wrapper_not_overwritten:
                continue  # not overwritten, fine

            # here means: overwritten
            if hasattr(method, "__wrapped__"):
                if method.__wrapped__ == supports:
                    if has_support:
                        continue  # needs support, has been wrapped

                    msg = (
                        f"Method {method_name} has been wrapped with supports "
                        "but is not allowed to. Has to handle all "
                        "arguments."
                    )
                    raise BasePDFSubclassingError(msg)
                if has_support:
                    msg = (
                        f"Method {method_name} has been overwritten and *has to* be "
                        "wrapped by `supports` decorator (don't forget () )"
                        "to call the decorator as it takes arguments"
                        ""
                    )
                    raise BasePDFSubclassingError(msg)
                if not has_support:
                    continue  # no support, has not been wrapped with
            elif not has_support:
                continue  # not wrapped, no support, need no

            # if we reach this points, somethings was implemented wrongly
            msg = (
                f"Method {method_name} has not been correctly wrapped with @supports "
                "OR has been wrapped but it should not be"
            )
            raise BasePDFSubclassingError(msg)

    @property
    def axes(self):
        return self.space.binning

    def to_binneddata(self, **kwargs) -> zfit.data.BinnedData:
        """Create an Asimov dataset as ``BinnedData`` using either ``counts`` (for extended) or ``rel_counts``.

        Args:
            **kwargs (): arguments to ``counts`` or ``rel_counts``.

        Returns:
            BinnedData: Binned data representing the Asimov dataset of this PDF.
        """
        values = self.values(**kwargs)
        return BinnedData.from_tensor(space=self.space, values=values)

    def to_hist(self, **kwargs):
        """Create an Asimov histogram as ``Hist`` using either ``counts`` (for extended) or ``rel_counts``.

        Args:
            **kwargs (): arguments to ``counts`` or ``rel_counts``.

        Returns:
            ``hist.Hist``: Histogram representing the Asimov dataset of this PDF.
        """
        return self.to_binneddata(**kwargs).to_hist()

    @property
    def space(self):
        return self._space

    def _set_yield(self, value):
        if self.is_extended:
            msg = f"Cannot extend {self}, is already extended."
            raise AlreadyExtendedPDFError(msg)
        value = convert_to_parameter(value)
        self.add_cache_deps(value)
        self._yield = value

    def _get_params(
        self,
        floating: bool | None,
        is_yield: bool | None,
        extract_independent: bool | None,
        *,
        autograd: bool | None = None,
    ) -> set[ZfitParameter]:
        params = super()._get_params(
            floating, is_yield=is_yield, extract_independent=extract_independent, autograd=autograd
        )

        if is_yield is not False:
            if self.is_extended:
                yield_params = extract_filter_params(
                    self.get_yield(),
                    floating=floating,
                    extract_independent=extract_independent,
                )
                # we care if it supports autograd or not
                if autograd is None or (autograd is False and "yield" not in self._autograd_params):
                    params = yield_params | params

                assert autograd is not True, "autograd should either be None or False, internal error"
            elif is_yield is True:
                msg = "PDF is not extended but only yield parameters were requested."
                raise NotExtendedPDFError(msg)
        return params

    def _convert_input_binned_x(self, x, none_is_space=None):
        if x is None and none_is_space:
            return self.space
        if isinstance(x, PlottableHistogram) and not isinstance(x, ZfitBinnedData):
            x = BinnedData.from_hist(x)
        if not isinstance(x, (ZfitBinnedData, ZfitSpace, ZfitUnbinnedData)):
            # TODO: should we allow spaces? Or what?
            try:
                x = Data.from_tensor(obs=self.obs, tensor=x)
            except Exception as error:
                raise TypeError(
                    f"Data to {self} has to be Binned Data, not {x}. (It can also be unbinned Data)"
                    + " but conversion to it failed (see also above) with the following error:"
                    + f" {error})"
                ) from error
        return x

    def _check_convert_norm(self, norm, none_is_error=False):
        if norm is None:
            norm = self.norm
        if norm is None:
            if none_is_error:
                msg = "norm cannot be None for this function."
                raise ValueError(msg)
        elif norm is not False and not isinstance(norm, ZfitSpace):
            msg = f"`norm` needs to be a binned ZfitSpace, not {norm}."
            raise TypeError(msg)
        elif norm is not False and not norm.is_binned:
            norm = norm.with_binning(self.space.binning)
        return norm

    def _check_convert_limits(self, limits):
        if limits is None:
            limits = self.space
        if not isinstance(limits, ZfitSpace):
            limits = convert_to_space(obs=self.obs, limits=limits)
        if isinstance(limits, ZfitSpace) and not limits.is_binned:
            limits = limits.with_binning(self.space.binning)
        return limits

    @_BinnedPDF_register_check_support(True)
    def _pdf(self, x, norm):
        return self._call_rel_counts(x, norm=norm) / np.prod(self.space.binning.widths, axis=0)

    @deprecated_norm_range
    def pdf(
        self,
        x: ztyping.XType,
        norm: ztyping.LimitsType = None,
        *,
        params: ztyping.ParamTypeInput = None,
    ) -> ztyping.XType:
        """Probability density function, evaluated at ``x`` or in the bins of ``x``

        Args:
            x: |@doc:binnedpdf.pdf.x| Values to evaluate the PDF at.
               If this is a ``ZfitBinnedData``-like object, a histogram of *densities*
               will be returned. If x is a ``ZfitUnbinnedData``-like object, the densities will be
               evaluated at the points of ``x``. |@docend:binnedpdf.pdf.x|
            norm: |@doc:pdf.pdf.norm| Normalization of the function.
               By default, this is the ``norm`` of the PDF (which by default is the same as
               the space of the PDF). |@docend:pdf.pdf.norm|
            params: |@doc:model.args.params| Mapping of the parameter names to the actual
               values. The parameter names refer to the names of the parameters,
               typically :py:class:`~zfit.Parameter`, that
               the model was _initialized_ with, not the name of the models
               parametrization. |@docend:model.args.params|

        Returns:
            ``Array-like``: probability density
        """

        # convert the input argument to a standardized form
        x = self._convert_input_binned_x(x, none_is_space=True)
        norm = self._check_convert_norm(norm, none_is_error=True)

        # sort it and remember the original sorting
        original_space = x if isinstance(x, ZfitSpace) else x.space
        x = x.with_obs(self.space.obs)  # we don't want to cut anything off -> only obs, not space

        # if it is unbinned, we get the binned version and gather the corresponding values
        is_unbinned = isinstance(x, ZfitUnbinnedData)
        binindices = None
        if is_unbinned:
            binindices = unbinned_to_binindex(x, self.space, flow=True)
            x = self.space
        with self._check_set_input_params(params=params):
            values = self._call_pdf(x, norm=norm)

        if binindices is not None:  # because we have the flow, so we need to make it here with pads
            padded_values = znp.pad(
                values,
                znp.ones((z._get_ndims(values), 2), dtype=znp.float64),
                mode="constant",
            )  # for overflow
            ordered_values = tf.gather_nd(padded_values, indices=binindices)
        else:
            ordered_values = move_axis_obs(self.space, original_space, values)[0]  # only use values, not variance
        return znp.asarray(ordered_values)

    @z.function(wraps="model_binned")
    def _call_pdf(self, x, norm):
        with suppress(SpecificFunctionNotImplemented):
            return self._auto_pdf(x, norm)
        with suppress(SpecificFunctionNotImplemented):
            return self._auto_rel_counts(x, norm=norm) / np.prod(self.space.binning.widths, axis=0)
        return self._fallback_pdf(x, norm=norm)

    def _auto_pdf(self, x, norm):
        try:
            pdf = self._pdf(x, norm=norm)
        except NormNotImplemented:
            unnormed_pdf = self._pdf(x, norm=False)
            normalization = self.normalization(norm)
            pdf = unnormed_pdf / normalization
        return pdf

    def _fallback_pdf(self, x, norm):
        values = self._call_unnormalized_pdf(x)
        if norm is not False:
            values = values / self.normalization(norm)
        return values

    def _unnormalized_pdf(self, x):  # noqa: ARG002
        raise SpecificFunctionNotImplemented

    def _call_unnormalized_pdf(self, x):
        return self._unnormalized_pdf(x)

    @_BinnedPDF_register_check_support(True)
    def _ext_pdf(self, x, norm):  # noqa: ARG002
        raise SpecificFunctionNotImplemented

    @deprecated_norm_range
    def ext_pdf(
        self,
        x: ztyping.XType,
        norm: ztyping.LimitsType = None,
        *,
        params: ztyping.ParamTypeInput = None,
    ) -> ztyping.XType:
        """Probability density function scaled by yield, evaluated at ``x`` or in the bins of ``x``

        Args:
            x: |@doc:binnedpdf.pdf.x| Values to evaluate the PDF at.
               If this is a ``ZfitBinnedData``-like object, a histogram of *densities*
               will be returned. If x is a ``ZfitUnbinnedData``-like object, the densities will be
               evaluated at the points of ``x``. |@docend:binnedpdf.pdf.x|
            norm: |@doc:pdf.pdf.norm| Normalization of the function.
               By default, this is the ``norm`` of the PDF (which by default is the same as
               the space of the PDF). |@docend:pdf.pdf.norm|
            params: |@doc:model.args.params| Mapping of the parameter names to the actual
               values. The parameter names refer to the names of the parameters,
               typically :py:class:`~zfit.Parameter`, that
               the model was _initialized_ with, not the name of the models
               parametrization. |@docend:model.args.params|

        Returns:
            |@doc:binnedpdf.out.problike| If the input was unbinned, it returns an array
               of shape (nevents,). If the input was binned, the dimensions and ordering of
               the axes corresponds to the input axes. |@docend:binnedpdf.out.problike|
        """
        if not self.is_extended:
            raise NotExtendedPDFError
        # convert the input argument to a standardized form
        x = self._convert_input_binned_x(x, none_is_space=True)
        norm = self._check_convert_norm(norm, none_is_error=True)
        # sort it and remember the original sorting
        original_space = x if isinstance(x, ZfitSpace) else x.space
        x = x.with_obs(self.space.obs)  # we don't want to cut anything off -> only obs, not space

        # if it is unbinned, we get the binned version and gather the corresponding values
        is_unbinned = isinstance(x, ZfitUnbinnedData)
        binindices = None
        if is_unbinned:
            binindices = unbinned_to_binindex(x, self.space, flow=True)
            x = self.space

        with self._check_set_input_params(params=params):
            values = self._call_ext_pdf(x, norm=norm)

        if binindices is not None:  # because we have the flow, so we need to make it here with pads
            padded_values = znp.pad(
                values,
                znp.ones((z._get_ndims(values), 2), dtype=znp.float64),
                mode="constant",
            )  # for overflow
            ordered_values = tf.gather_nd(padded_values, indices=binindices)
        else:
            ordered_values = move_axis_obs(self.space, original_space, values)[0]  # only use values, not variance
        return znp.asarray(ordered_values)

    @z.function(wraps="model_binned")
    def _call_ext_pdf(self, x, norm):
        with suppress(SpecificFunctionNotImplemented):
            return self._auto_ext_pdf(x, norm)
        with suppress(SpecificFunctionNotImplemented):
            return self._auto_counts(x, norm=norm) / np.prod(self.space.binning.widths, axis=0)
        return self._fallback_ext_pdf(x, norm=norm)

    def _auto_ext_pdf(self, x, norm):
        try:
            pdf = self._ext_pdf(x, norm=norm)
        except NormNotImplemented:
            unnormed_pdf = self._ext_pdf(x, norm=False)
            normalization = self.ext_normalization(norm)
            pdf = unnormed_pdf / normalization
        return pdf

    def _fallback_ext_pdf(self, x, norm):
        values = self._call_pdf(x, norm=norm)
        return values * self.get_yield()

    @_BinnedPDF_register_check_support(True)
    def _log_pdf(self, x, *, norm, params=None):  # noqa: ARG002
        raise SpecificFunctionNotImplemented

    def log_pdf(
        self,
        x: ztyping.XType,
        norm: ztyping.LimitsType = None,
        *,
        params: ztyping.ParamTypeInput = None,
    ) -> ztyping.XType:
        """Log probability density function, evaluated at ``x`` or in the bins of ``x``

        Args:
            x: |@doc:binnedpdf.pdf.x| Values to evaluate the PDF at.
               If this is a ``ZfitBinnedData``-like object, a histogram of *densities*
               will be returned. If x is a ``ZfitUnbinnedData``-like object, the densities will be
               evaluated at the points of ``x``. |@docend:binnedpdf.pdf.x|
            norm: |@doc:pdf.pdf.norm| Normalization of the function.
               By default, this is the ``norm`` of the PDF (which by default is the same as
               the space of the PDF). |@docend:pdf.pdf.norm|
            params: |@doc:model.args.params| Mapping of the parameter names to the actual
               values. The parameter names refer to the names of the parameters,
               typically :py:class:`~zfit.Parameter`, that
               the model was _initialized_ with, not the name of the models
               parametrization. |@docend:model.args.params|

        Returns:
            |@doc:binnedpdf.out.problike| If the input was unbinned, it returns an array
               of shape (nevents,). If the input was binned, the dimensions and ordering of
               the axes corresponds to the input axes. |@docend:binnedpdf.out.problike|
        """
        # convert the input argument to a standardized form
        x = self._convert_input_binned_x(x, none_is_space=True)
        norm = self._check_convert_norm(norm, none_is_error=True)
        # sort it and remember the original sorting
        original_space = x if isinstance(x, ZfitSpace) else x.space
        x = x.with_obs(self.space.obs)  # we don't want to cut anything off -> only obs, not space

        # if it is unbinned, we get the binned version and gather the corresponding values
        is_unbinned = isinstance(x, ZfitUnbinnedData)
        binindices = None
        if is_unbinned:
            binindices = unbinned_to_binindex(x, self.space, flow=True)
            x = self.space

        with self._check_set_input_params(params=params):
            values = self._call_log_pdf(x, norm=norm)

        if binindices is not None:  # because we have the flow, so we need to make it here with pads
            padded_values = znp.pad(
                values,
                znp.ones((z._get_ndims(values), 2), dtype=znp.float64),
                mode="constant",
            )  # for overflow
            ordered_values = tf.gather_nd(padded_values, indices=binindices)
        else:
            ordered_values = move_axis_obs(self.space, original_space, values)[0]  # only use values, not variance
        return znp.asarray(ordered_values)

    @z.function(wraps="model_binned")
    def _call_log_pdf(self, x, norm):
        with suppress(SpecificFunctionNotImplemented):
            return self._auto_log_pdf(x, norm)
        with suppress(SpecificFunctionNotImplemented):
            return znp.log(self._call_pdf(x, norm=norm))
        return self._fallback_log_pdf(x, norm=norm)

    def _auto_log_pdf(self, x, norm):
        try:
            pdf = self._log_pdf(x, norm=norm)
        except NormNotImplemented:
            unnormed_pdf = self._log_pdf(x, norm=False)
            log_normalization = self.log_normalization(norm)
            pdf = unnormed_pdf - log_normalization
        return pdf

    def _fallback_log_pdf(self, x, norm):
        values = self._call_pdf(x, norm=norm)
        return znp.log(values)

    @_BinnedPDF_register_check_support(True)
    def _ext_log_pdf(self, x, *, norm, params=None):  # noqa: ARG002
        raise SpecificFunctionNotImplemented

    def ext_log_pdf(
        self,
        x: ztyping.XType,
        norm: ztyping.LimitsType = None,
        *,
        params: ztyping.ParamTypeInput = None,
    ) -> ztyping.XType:
        """Log probability density function scaled by yield, evaluated at ``x`` or in the bins of ``x``

        Args:
            x: |@doc:binnedpdf.pdf.x| Values to evaluate the PDF at.
               If this is a ``ZfitBinnedData``-like object, a histogram of *densities*
               will be returned. If x is a ``ZfitUnbinnedData``-like object, the densities will be
               evaluated at the points of ``x``. |@docend:binnedpdf.pdf.x|
            norm: |@doc:pdf.pdf.norm| Normalization of the function.
               By default, this is the ``norm`` of the PDF (which by default is the same as
               the space of the PDF). |@docend:pdf.pdf.norm|
            params: |@doc:model.args.params| Mapping of the parameter names to the actual
               values. The parameter names refer to the names of the parameters,
               typically :py:class:`~zfit.Parameter`, that
               the model was _initialized_ with, not the name of the models
               parametrization. |@docend:model.args.params|

        Returns:
            |@doc:binnedpdf.out.problike| If the input was unbinned, it returns an array
               of shape (nevents,). If the input was binned, the dimensions and ordering of
               the axes corresponds to the input axes. |@docend:binnedpdf.out.problike|
        """
        if not self.is_extended:
            raise NotExtendedPDFError
        # convert the input argument to a standardized form
        x = self._convert_input_binned_x(x, none_is_space=True)
        norm = self._check_convert_norm(norm, none_is_error=True)
        # sort it and remember the original sorting
        original_space = x if isinstance(x, ZfitSpace) else x.space
        x = x.with_obs(self.space.obs)  # we don't want to cut anything off -> only obs, not space

        # if it is unbinned, we get the binned version and gather the corresponding values
        is_unbinned = isinstance(x, ZfitUnbinnedData)
        binindices = None
        if is_unbinned:
            binindices = unbinned_to_binindex(x, self.space, flow=True)
            x = self.space

        with self._check_set_input_params(params=params):
            values = self._call_ext_log_pdf(x, norm=norm)

        if binindices is not None:  # because we have the flow, so we need to make it here with pads
            padded_values = znp.pad(
                values,
                znp.ones((z._get_ndims(values), 2), dtype=znp.float64),
                mode="constant",
            )  # for overflow
            ordered_values = tf.gather_nd(padded_values, indices=binindices)
        else:
            ordered_values = move_axis_obs(self.space, original_space, values)[0]  # only use values, not variance
        return znp.asarray(ordered_values)

    def _call_ext_log_pdf(self, x, norm):
        with suppress(SpecificFunctionNotImplemented):
            return self._auto_ext_log_pdf(x, norm)
        with suppress(SpecificFunctionNotImplemented):
            return znp.log(self._call_ext_pdf(x, norm=norm))
        return self._fallback_ext_log_pdf(x, norm=norm)

    def _auto_ext_log_pdf(self, x, norm):
        try:
            pdf = self._ext_log_pdf(x, norm=norm)
        except NormNotImplemented:
            unnormed_pdf = self._ext_log_pdf(x, norm=False)
            log_normalization = znp.log(self.ext_normalization(norm))
            pdf = unnormed_pdf - log_normalization
        return pdf

    def _fallback_ext_log_pdf(self, x, norm):
        values = self._call_ext_pdf(x, norm=norm)
        return znp.log(values)

    @_BinnedPDF_register_check_support(True)
    def _log_normalization(self, norm, *, params=None):  # noqa: ARG002
        raise SpecificFunctionNotImplemented

    @deprecated_args(None, "Use `norm` instead.", "limits")
    def log_normalization(self, norm, *, params=None, options=None, limits=None) -> ztyping.NumericalTypeReturn:
        """Normalization of the PDF. For a binned PDF, this is the sum over the counts or the integral over the density.

        Args:
            norm: |@doc:pdf.normalization.norm| Normalization of the function.
                By default, this is the ``norm`` of the PDF (which by default is the same as
                the space of the PDF). |@docend:pdf.normalization.norm|
            options:   |@doc:pdf.normalization.options| Additional options for the normalization.
                Currently supported options are:
                - type: one of (``bins``)
                  This hints that bins are integrated. A method that is vectorizable, non-dynamic and
                  therefore less suitable for complicated functions is chosen. |@docend:pdf.normalization.options|

            params: |@doc:model.args.params| Mapping of the parameter names to the actual
               values. The parameter names refer to the names of the parameters,
               typically :py:class:`~zfit.Parameter`, that
               the model was _initialized_ with, not the name of the models
               parametrization. |@docend:model.args.params|

        Returns:
        """
        if limits is not None:
            norm = limits
        if options is None:
            options = {}
        norm = self._check_convert_norm(norm, none_is_error=True)
        with self._check_set_input_params(params=params):
            return self._call_log_normalization(norm, options=options)

    def _call_log_normalization(self, norm, *, options):
        with suppress(SpecificFunctionNotImplemented):
            return self._log_normalization(norm, options=options)

        # fallback
        return znp.log(self._call_integrate(norm, norm=False, options=None))

    @_BinnedPDF_register_check_support(True)
    def _normalization(self, limits, *, options, params=None):  # noqa: ARG002
        raise SpecificFunctionNotImplemented

    def normalization(self, norm=None, *, params=None, options=None, limits=None) -> ztyping.NumericalTypeReturn:
        """Normalization of the PDF. For a binned PDF, this is the sum over the counts or the integral over the density.

        Args:
            norm: |@doc:pdf.normalization.norm| Normalization of the function.
                By default, this is the ``norm`` of the PDF (which by default is the same as
                the space of the PDF). |@docend:pdf.normalization.norm|
            options:   |@doc:pdf.normalization.options| Additional options for the normalization.
                Currently supported options are:
                - type: one of (``bins``)
                  This hints that bins are integrated. A method that is vectorizable, non-dynamic and
                  therefore less suitable for complicated functions is chosen. |@docend:pdf.normalization.options|

            params: |@doc:model.args.params| Mapping of the parameter names to the actual
               values. The parameter names refer to the names of the parameters,
               typically :py:class:`~zfit.Parameter`, that
               the model was _initialized_ with, not the name of the models
               parametrization. |@docend:model.args.params|

        Returns:
        """
        if limits is not None:
            msg = "Use `norm` instead of `limits`."
            raise BreakingAPIChangeError(msg)
        if options is None:
            options = {}
        norm = self._check_convert_norm(norm)
        with self._check_set_input_params(params=params):
            return self._call_normalization(norm, options=options)

    def _call_normalization(self, norm, *, options):
        with suppress(SpecificFunctionNotImplemented):
            return self._normalization(norm, options=options)

        # fallback
        return self._call_integrate(norm, norm=False, options=None)

    @_BinnedPDF_register_check_support(True)
    def _integrate(self, limits, norm, options, params=None):  # noqa: ARG002
        raise SpecificFunctionNotImplemented

    def integrate(
        self,
        limits: ztyping.LimitsType,
        norm: ztyping.LimitsType = None,
        *,
        options=None,
        params: ztyping.ParamTypeInput = None,
    ) -> ztyping.XType:
        """Integral of the PDF, the sum over all the bins normalized to the total integral.

        Args:
            limits: |@doc:pdf.integrate.limits| Limits of the integration. |@docend:pdf.integrate.limits|
            norm: |@doc:pdf.integrate.norm| Normalization of the integration.
               By default, this is the same as the default space of the PDF.
               ``False`` means no normalization and returns the unnormed integral. |@docend:pdf.integrate.norm|
            options: |@doc:pdf.integrate.options| Options for the integration.
               Additional options for the integration. Currently supported options are:
               - type: one of (``bins``)
                 This hints that bins are integrated. A method that is vectorizable,
                 non-dynamic and therefore less suitable for complicated functions is chosen. |@docend:pdf.integrate.options|
            params: |@doc:model.args.params| Mapping of the parameter names to the actual
               values. The parameter names refer to the names of the parameters,
               typically :py:class:`~zfit.Parameter`, that
               the model was _initialized_ with, not the name of the models
               parametrization. |@docend:model.args.params|

        Returns:
            Scalar integration value.
        """
        if options is None:
            options = {}
        norm = self._check_convert_norm(norm)
        limits = self._check_convert_limits(limits)
        with self._check_set_input_params(params=params):
            return self._call_integrate(limits, norm, options)

    @z.function(wraps="model_binned")
    def _call_integrate(self, limits, norm, options=None):
        if options is None:
            options = {}
        with suppress(SpecificFunctionNotImplemented):
            return self._auto_integrate(limits, norm, options)
        return self._fallback_integrate(limits, norm, options)

    def _fallback_integrate(self, limits, norm, options):
        del options  # not used
        bincounts = self._call_rel_counts(limits, norm=norm)  # TODO: fake data? not to integrate limits?
        edges = limits.binning.edges
        return binned_rect_integration(counts=bincounts, edges=edges, limits=limits)  # TODO: check integral, correct?

    def _auto_integrate(self, limits, norm, options):
        try:
            integral = self._integrate(limits=limits, norm=norm, options=options)
        except NormNotImplemented:
            unnormalized_integral = self._auto_integrate(limits=limits, norm=False, options=options)
            normalization = self.normalization(norm, options=options)
            integral = unnormalized_integral / normalization
        except MultipleLimitsNotImplemented:
            integrals = []  # TODO: map?
            for sub_limits in limits:
                integrals.append(self._auto_integrate(limits=sub_limits, norm=norm, options=options))
            integral = z.reduce_sum(integrals, axis=0)  # TODO: remove stack?
        return integral

    @deprecated_norm_range
    def ext_integrate(
        self,
        limits: ztyping.LimitsType,
        norm: ztyping.LimitsType = None,
        *,
        options=None,
        params: ztyping.ParamTypeInput = None,
    ) -> ztyping.XType:
        """Extended integral of the PDF, i.e. the expected counts or the integral scaled by the yield.

        Args:
            limits: |@doc:pdf.integrate.limits| Limits of the integration. |@docend:pdf.integrate.limits|
            norm: |@doc:pdf.integrate.norm| Normalization of the integration.
               By default, this is the same as the default space of the PDF.
               ``False`` means no normalization and returns the unnormed integral. |@docend:pdf.integrate.norm|
            options: |@doc:pdf.integrate.options| Options for the integration.
               Additional options for the integration. Currently supported options are:
               - type: one of (``bins``)
                 This hints that bins are integrated. A method that is vectorizable,
                 non-dynamic and therefore less suitable for complicated functions is chosen. |@docend:pdf.integrate.options|
            params: |@doc:model.args.params| Mapping of the parameter names to the actual
               values. The parameter names refer to the names of the parameters,
               typically :py:class:`~zfit.Parameter`, that
               the model was _initialized_ with, not the name of the models
               parametrization. |@docend:model.args.params|

        Returns:
            Scalar integration value.
        """
        if not self.is_extended:
            raise NotExtendedPDFError
        if options is None:
            options = {}
        norm = self._check_convert_norm(norm)
        limits = self._check_convert_limits(limits)
        with self._check_set_input_params(params=params):
            return self._call_ext_integrate(limits, norm, options=options)

    @z.function(wraps="model_binned")
    def _call_ext_integrate(self, limits, norm, *, options):
        with suppress(SpecificFunctionNotImplemented):
            return self._auto_ext_integrate(limits, norm, options=options)
        return self._fallback_ext_integrate(limits, norm, options=options)

    def _fallback_ext_integrate(self, limits, norm, *, options):  # TODO: rather use pdf?
        del options  # not used
        bincounts = self._call_counts(limits, norm=norm)  # TODO: fake data? not to integrate limits?
        edges = limits.binning.edges
        return binned_rect_integration(counts=bincounts, edges=edges, limits=limits)

    def _auto_ext_integrate(self, limits, norm, *, options):
        try:
            integral = self._ext_integrate(limits=limits, norm=norm, options=options)
        except NormNotImplemented:
            unnormalized_integral = self._auto_ext_integrate(limits=limits, norm=False, options=options)
            normalization = self.ext_normalization(norm, options=options)
            integral = unnormalized_integral / normalization
        except MultipleLimitsNotImplemented:
            integrals = []  # TODO: map?
            for sub_limits in limits:
                integrals.append(self._auto_integrate(limits=sub_limits, norm=norm, options=options))
            integral = z.reduce_sum(integrals, axis=0)  # TODO: remove stack?
        return integral

    @_BinnedPDF_register_check_support(True)
    def _ext_integrate(self, limits, norm, *, options, params=None):  # noqa: ARG002
        raise SpecificFunctionNotImplemented

    def create_sampler(
        self,
        n: ztyping.nSamplingTypeIn = None,
        limits: ztyping.LimitsType = None,
        *,
        params: ztyping.ParamTypeInput = None,
        fixed_params: bool | list[ZfitParameter] | tuple[ZfitParameter] = True,  # todo: use params instead only?
    ) -> BinnedSamplerData:
        """Create a :py:class:`SamplerData` that acts as `Data` but can be resampled, also with changed parameters and
        n.

            If `limits` is not specified, `space` is used (if the space contains limits).
            If `n` is None and the model is an extended pdf, 'extended' is used by default.


        Args:
            n: The number of samples to be generated. Can be a Tensor that will be
                or a valid string. Currently implemented:

                    - 'extended': samples `poisson(yield)` from each pdf that is extended.

            limits: From which space to sample.
            params: |@doc:model.args.params| Mapping of the parameter names to the actual
               values. The parameter names refer to the names of the parameters,
               typically :py:class:`~zfit.Parameter`, that
               the model was _initialized_ with, not the name of the models
               parametrization. |@docend:model.args.params|
            fixed_params: A list of `Parameters` that will be fixed during several `resample` calls.
                If True, all are fixed, if False, all are floating. If a :py:class:`~zfit.Parameter` is not fixed and
                its
                value gets updated (e.g. by a `Parameter.set_value()` call), this will be reflected in
                `resample`. If fixed, the Parameter will still have the same value as the `SamplerData` has
                been created with when it resamples.

        Returns:
            :py:class:`~zfit.core.data.BinnedSampler`

        Raises:
            NotExtendedPDFError: if 'extended' is chosen (implicitly by default or explicitly) as an
                option for `n` but the pdf itself is not extended.
            ValueError: if n is an invalid string option.
            InvalidArgumentError: if n is not specified and pdf is not extended.
        """

        if n is None:
            if self.is_extended:
                n = znp.random.poisson(self.get_yield(), size=1)
            else:
                msg = f"n cannot be None for sampling of {self} or needs to be extended."
                raise ValueError(msg)
        limits = self._check_convert_limits(limits)

        if fixed_params is not None:
            if params is not None:
                msg = "Cannot specify both `fixed_params` and `params`. Use `params` only."
                raise ValueError(msg)
            if fixed_params is False:
                msg = (
                    "`fixed_params` has been removed, the sampler will always sample from the parameters at the time of the creation/given to the creator"
                    " _or_ by giving params to the `resample` method."
                )
                raise BreakingAPIChangeError(msg)
            if fixed_params is True:
                fixed_params = None  # default behavior is to catch all anyway
            if isinstance(fixed_params, (list, tuple)):
                fixed_params = {param.name: znp.asarray(param) for param in fixed_params}
            params = fixed_params

        # legacy end
        params = self._check_convert_input_paramvalues(params=params)
        params = {
            p.name: params[pname] if (pname := p) in params or (pname := p.name) in params else p.value()
            for p in self.get_params(floating=None, is_yield=None)
        }

        def sample_and_variances_func(n, params, *, limits=limits):
            sample = self.sample(n=n, limits=limits, params=params)
            return sample.values(), sample.variances()

        return BinnedSamplerData.from_sampler(
            sample_and_variances_func=sample_and_variances_func,
            n=n,
            obs=limits,
            params=params,
        )

    @z.function(wraps="sampler")
    def _create_sampler_tensor(self, limits, n):
        return self._call_sample(n=n, limits=limits)

    def sample(
        self, n: int | None = None, limits: ztyping.LimitsType = None, *, params: ztyping.ParamTypeInput = None
    ) -> ZfitBinnedData:
        """Draw a random binned sample from the PDF.

        Args:
            n: |@doc:pdf.sample.n| Number of samples to draw.
               For an extended PDF, the argument is optional and will be the
               poisson-fluctuated expected number of events, i.e. the yield. |@docend:pdf.sample.n|
            limits: |@doc:pdf.sample.limits| Limits of the sampling.
               By default, this is the same as the default space of the PDF. |@docend:pdf.sample.limits|
            params: |@doc:model.args.params| Mapping of the parameter names to the actual
               values. The parameter names refer to the names of the parameters,
               typically :py:class:`~zfit.Parameter`, that
               the model was _initialized_ with, not the name of the models
               parametrization. |@docend:model.args.params|

        Returns:
            ``ZfitBinnedData``: Sampled dataset
        """
        if n is None:
            if self.is_extended:
                n = znp.random.poisson(self.get_yield(), size=1)
            else:
                msg = f"n cannot be None for sampling of {self} or needs to be extended."
                raise ValueError(msg)
        original_limits = limits
        limits = self._check_convert_limits(limits)
        with self._check_set_input_params(params=params):
            values = self._call_sample(n, limits)
        if not isinstance(values, ZfitBinnedData):
            values = BinnedData.from_tensor(space=limits, values=values, variances=None)
        if isinstance(original_limits, ZfitSpace):
            values = values.with_obs(original_limits)
        return values

    @z.function(wraps="sample")
    def _call_sample(self, n, limits):
        with suppress(SpecificFunctionNotImplemented):
            self._sample(n, limits)
        return self._fallback_sample(n, limits)

    def _fallback_sample(self, n, limits):
        if limits != self.space:
            msg = (
                "limits different from the default are not yet available."
                " Please open an issue if you need this:"
                " https://github.com/zfit/zfit/issues/new/choose"
            )
            raise WorkInProgressError(msg)
        probs = self.rel_counts(limits)
        return z.random.counts_multinomial(n, probs=probs, dtype=znp.float64)

    # ZfitMinimalHist implementation
    def values(self, *, var=None, params=None):
        """Histogram values that are either the counts or the normalized counts.

        If the PDF is extended, the counts are returned, otherwise the normalized counts are returned.

        Args:
            params: |@doc:model.args.params| Mapping of the parameter names to the actual
               values. The parameter names refer to the names of the parameters,
               typically :py:class:`~zfit.Parameter`, that
               the model was _initialized_ with, not the name of the models
               parametrization. |@docend:model.args.params|

        Returns:
            ``ZfitBinnedData``: Histogram values
        """
        if var is not None:
            msg = "var argument for `values` is not supported in V1"
            raise RuntimeError(msg)
        if self.is_extended:
            return self.counts(var, params=params)
        else:
            return self.rel_counts(var, params=params)

    def update_integration_options(self, *_, **__):
        msg = "Integration options not available for BinnedPDF"
        raise RuntimeError(msg)

    def as_func(self, norm: ztyping.LimitsType = False):  # noqa: ARG002
        msg = "as_func not yet available for BinnedPDF"
        raise WorkInProgressError(msg)

    @property
    def is_extended(self) -> bool:
        return self._yield is not None

    def set_norm(self, norm):  # noqa: ARG002
        msg = "set_norm should not be used anymore. Create a new PDF with the desired normalization."
        raise RuntimeError(msg)

    def create_extended(self, yield_: ztyping.ParamTypeInput) -> ZfitPDF:  # noqa: ARG002
        msg = "create_extended not available for BinnedPDF. Use `extended` in the initialization" " instead."
        raise WorkInProgressError(msg)

    def get_yield(self) -> ZfitParameter | None:
        if not self.is_extended:
            raise NotExtendedPDFError
        return self._yield

    @classmethod
    def register_analytic_integral(
        cls,
        func: Callable,  # noqa: ARG003
        limits: ztyping.LimitsType = None,  # noqa: ARG003
        priority: int = 50,  # noqa: ARG003
        *,
        supports_norm: bool = False,  # noqa: ARG003
        supports_multiple_limits: bool = False,  # noqa: ARG003
    ):
        msg = "analytic integral not available for BinnedPDF"
        raise RuntimeError(msg)

    @deprecated_norm_range
    def partial_integrate(
        self,
        x: ztyping.XType,  # noqa: ARG002
        limits: ztyping.LimitsType,  # noqa: ARG002
        *,
        norm=None,  # noqa: ARG002
        options=None,  # noqa: ARG002
    ) -> ztyping.XType:
        msg = "partial_integrate not yet available for BinnedPDF"
        raise WorkInProgressError(msg)

    @classmethod
    def register_inverse_analytic_integral(cls, func: Callable):  # noqa: ARG003
        msg = "inverse analytic integral not available for BinnedPDF. It's a histogram, it's already " "'analytic'"
        raise RuntimeError(msg)

    @_BinnedPDF_register_check_support(True)
    def _sample(self, n, limits, *, params=None):  # noqa: ARG002
        raise SpecificFunctionNotImplemented

    def _copy(self, deep, name, overwrite_params):  # noqa: ARG002
        raise WorkInProgressError

    # factor out with unbinned pdf

    @property
    def norm(self):
        norm = self._norm
        if norm is None:
            norm = self.space
        return norm

    @property
    @deprecated(None, "use `norm` instead.")
    def norm_range(self):
        return self.norm

    # TODO: factor out with unbinned pdf

    def _convert_sort_space(
        self,
        obs: ztyping.ObsTypeInput | ztyping.LimitsTypeInput = None,
        axes: ztyping.AxesTypeInput = None,
        limits: ztyping.LimitsTypeInput = None,
    ) -> ZfitSpace | None:
        """Convert the inputs (using eventually ``obs``, ``axes``) to :py:class:`~zfit.ZfitSpace` and sort them
        according to own ``obs``.

        Args:
            obs:
            axes:
            limits:

        Returns:
        """
        if obs is None:  # for simple limits to convert them
            obs = self.obs
        elif not set(obs).intersection(self.obs):
            msg = "The given space {obs} is not compatible with the obs of the pdfs{self.obs};" " they are disjoint."
            raise SpaceIncompatibleError(msg)
        space = convert_to_space(obs=obs, axes=axes, limits=limits)

        if self.space is not None:  # e.g. not the first call
            space = space.with_coords(self.space, allow_superset=True, allow_subset=True)
        return space

    def counts(
        self,
        x: ztyping.BinnedDataInputType = None,
        norm: ztyping.NormInputType = None,
        *,
        params: ztyping.ParamTypeInput = None,
    ) -> ZfitBinnedData:
        """Calculate the number of events in each bin.

        This is the integrals of the PDF in each bin.

        Args:
            x: |@doc:pdf.binned.counts.x| Data for the binned PDF.
               The returned counts correspond to the binned axis in ``x``. |@docend:pdf.binned.counts.x|
            norm: |@doc:pdf.binned.counts.norm| Normalization of the counts.
               This normalizes the counts so that the actual sum of all counts is
               equal to the yield. |@docend:pdf.binned.counts.norm|
            params: |@doc:model.args.params| Mapping of the parameter names to the actual
               values. The parameter names refer to the names of the parameters,
               typically :py:class:`~zfit.Parameter`, that
               the model was _initialized_ with, not the name of the models
               parametrization. |@docend:model.args.params|

        Returns:
            ZfitBinnedData: A histogram with the number of events in each bin.
        """

        if not self.is_extended:
            raise NotExtendedPDFError
        # convert the input argument to a standardized form
        x = self._convert_input_binned_x(x, none_is_space=True)
        norm = self._check_convert_norm(norm, none_is_error=True)
        # sort it and remember the original sorting
        original_space = x if isinstance(x, ZfitSpace) else x.space
        x = x.with_obs(self.space.obs)  # we don't want to cut anything off -> only obs, not space

        # if it is unbinned, we get the binned version and gather the corresponding values
        is_unbinned = isinstance(x, ZfitUnbinnedData)
        binindices = None
        if is_unbinned:
            binindices = unbinned_to_binindex(x, self.space, flow=True)
            x = self.space

        with self._check_set_input_params(params=params):
            values = self._call_counts(x, norm)

        if binindices is not None:  # because we have the flow, so we need to make it here with pads
            padded_values = znp.pad(
                values,
                znp.ones((z._get_ndims(values), 2), dtype=znp.float64),
                mode="constant",
            )  # for overflow
            ordered_values = tf.gather_nd(padded_values, indices=binindices)
        else:
            ordered_values = move_axis_obs(self.space, original_space, values)[0]  # only use values, not variance
        return znp.asarray(ordered_values)

    @z.function(wraps="model_binned")
    def _call_counts(self, x, norm):
        with suppress(SpecificFunctionNotImplemented):
            return self._auto_counts(x, norm)
        return self._fallback_counts(x, norm)

    def ext_normalization(
        self, norm: ztyping.NormInputType, *, options=None, params: ztyping.ParamTypeInput = None
    ) -> ztyping.NumericalTypeReturn:
        """Calculate the normalization of the extended PDF, for a binned PDF this is the sum of the counts.

        Args:
            norm: |@doc:pdf.normalization.norm| Normalization of the function.
                By default, this is the ``norm`` of the PDF (which by default is the same as
                the space of the PDF). |@docend:pdf.normalization.norm|
            options:  |@doc:pdf.normalization.options| Additional options for the normalization.
                Currently supported options are:
                - type: one of (``bins``)
                    This hints that bins are integrated. A method that is vectorizable, non-dynamic and
                    therefore less suitable for complicated functions is chosen. |@docend:pdf.normalization.options|
            params: |@doc:model.args.params| Mapping of the parameter names to the actual
               values. The parameter names refer to the names of the parameters,
               typically :py:class:`~zfit.Parameter`, that
               the model was _initialized_ with, not the name of the models
               parametrization. |@docend:model.args.params|

        Returns:
            Scalar-like: The normalization of the extended PDF.
        """
        if not self.is_extended:
            raise NotExtendedPDFError
        if options is None:
            options = {}
        norm = self._check_convert_norm(norm, none_is_error=True)
        with self._check_set_input_params(params=params):
            return self._call_ext_normalization(norm, options=options)

    def _call_ext_normalization(self, norm, *, options):
        with suppress(SpecificFunctionNotImplemented):
            return self._ext_normalization(norm, options=options)
        # fallback
        return self.normalization(norm, options=options) / self.get_yield()

    def _ext_normalization(self, norm, *, options):  # noqa: ARG002
        raise SpecificFunctionNotImplemented

    def _auto_counts(self, x, norm):
        try:
            counts = self._counts(x, norm=norm)
        except NormNotImplemented:
            unnormed_counts = self._counts(x, norm=False)
            normalization = self.normalization(norm)
            counts = unnormed_counts / normalization
        return counts

    def _fallback_counts(self, x, norm):
        return self._auto_rel_counts(x, norm) * self.get_yield()

    @_BinnedPDF_register_check_support(True)
    def _counts(self, x, norm):  # noqa: ARG002
        raise SpecificFunctionNotImplemented

    def rel_counts(
        self,
        x: ztyping.BinnedDataInputType = None,
        *,
        norm: ztyping.NormInputType = None,
        params: ztyping.ParamTypeInput = None,
    ) -> ZfitBinnedData:
        """Calculate the relative number of events in each bin.

        This is the integrals of the PDF in each bin divided by the integral of the PDF over the whole space.
        It is *not* equal to the density but rather a histogram scaled to 1.

        Args:
            x: |@doc:pdf.binned.counts.x| Data for the binned PDF.
               The returned counts correspond to the binned axis in ``x``. |@docend:pdf.binned.counts.x|
            norm: |@doc:pdf.binned.counts.norm| Normalization of the counts.
               This normalizes the counts so that the actual sum of all counts is
               equal to the yield. |@docend:pdf.binned.counts.norm|
            params: |@doc:model.args.params| Mapping of the parameter names to the actual
               values. The parameter names refer to the names of the parameters,
               typically :py:class:`~zfit.Parameter`, that
               the model was _initialized_ with, not the name of the models
               parametrization. |@docend:model.args.params|

        Returns:
            ZfitBinnedData: A histogram with the relative number of events in each bin.
        """
        # convert the input argument to a standardized form
        x = self._convert_input_binned_x(x, none_is_space=True)
        norm = self._check_convert_norm(norm, none_is_error=True)
        # sort it and remember the original sorting
        original_space = x if isinstance(x, ZfitSpace) else x.space
        x = x.with_obs(self.space.obs)  # we don't want to cut anything off -> only obs, not space

        # if it is unbinned, we get the binned version and gather the corresponding values
        is_unbinned = isinstance(x, ZfitUnbinnedData)
        binindices = None
        if is_unbinned:
            binindices = unbinned_to_binindex(x, self.space, flow=True)
            x = self.space

        with self._check_set_input_params(params=params):
            values = self._call_rel_counts(x, norm)

        if binindices is not None:  # because we have the flow, so we need to make it here with pads
            padded_values = znp.pad(
                values,
                znp.ones((z._get_ndims(values), 2), dtype=znp.float64),
                mode="constant",
            )  # for overflow
            ordered_values = tf.gather_nd(padded_values, indices=binindices)
        else:
            ordered_values = move_axis_obs(self.space, original_space, values)[0]  # only use values, not variance
        return znp.asarray(ordered_values)

    @z.function(wraps="model_binned")
    def _call_rel_counts(self, x, norm):
        with suppress(SpecificFunctionNotImplemented):
            return self._auto_rel_counts(x, norm=norm)
        with suppress(SpecificFunctionNotImplemented):
            return self._auto_counts(x, norm=norm) / self.get_yield()
        return self._fallback_rel_counts(x, norm)

    @_BinnedPDF_register_check_support(True)
    def _rel_counts(self, x, norm, *, params=None):  # noqa: ARG002
        raise SpecificFunctionNotImplemented

    def _auto_rel_counts(self, x, norm):
        try:
            rel_counts = self._rel_counts(x, norm=norm)
        except NormNotImplemented:
            unnormed_rel_counts = self._rel_counts(x, norm=False)
            normalization = self.normalization(norm)
            rel_counts = unnormed_rel_counts / normalization
        return rel_counts

    def _fallback_rel_counts(self, x, norm):
        density = self._call_pdf(x, norm)
        return density * np.prod(self.space.binning.widths, axis=0)

    def set_norm_range(self):
        msg = "set_norm_range is removed and should not be used anymore."
        raise RuntimeError(msg)

    def to_binned(self, space, *, extended=None, norm=None):
        """Convert the PDF to a binned PDF, returns self.

        For compatibility with unbinned PDFs.
        """
        if isinstance(space, ZfitBinning) and space != self.space.binning:
            msg = "The binning of the PDF and the binning of the space must be equal."
            raise ValueError(msg)
        if space != self.space:
            msg = f"Space must be the same as the PDF's space, as {self} is already a binned PDF."
            raise ValueError(msg)
        if extended is not None:
            msg = "extended is not implemented yet. Create an extended PDF manually."
            raise WorkInProgressError(msg)
        if norm is not None:
            msg = "norm is not implemented yet. Create a pdf with a different norm range manually."
            raise WorkInProgressError(msg)
        return self

    def to_unbinned(self):
        """Convert the PDF to an unbinned PDF."""
        from zfit.models.unbinnedpdf import UnbinnedFromBinnedPDF

        return UnbinnedFromBinnedPDF(self, self.space.with_binning(None))


BaseBinnedPDFV1 = BaseBinnedPDF  # alias, backwards compatibility


def binned_rect_integration(
    *,
    limits: ZfitSpace,
    edges: Iterable[znp.array] | znp.array,
    counts: znp.array | None = None,
    density: znp.array | None = None,
    axis: Iterable[int] | int | None = None,
) -> znp.array:
    """Integrate a histogram over *limits*.

    This integrator does take into account that limits do not match the edges.

    Args:
        limits: Limits to integrate over. A possible binning is ignored.
        edges: The edges per axis. They should have the shape ``(1,..., 1, n, 1, ..., 1)``, where n is the *ith* axis.
            ``ZfitBinning`` provides this format on the ``edges` attribute.
        counts: Counts of the histogram. This is what most histograms have and is equal to the density multiplied by
            the binwidth.
            Exactly one of counts or density has to be provided.
        density: The density of a histogram is the bincount divided by the binwidth.
            Exactly one of counts or density has to be provided.
        axis: Which axes to integrate over. Defaults to all.

    Returns:
        Integral with shape corresponding to the non-integrated axes (or a scalar in case of all axes integrated).
    """
    edges = convert_to_container(edges)
    if not isinstance(limits, ZfitSpace):
        msg = f"limits has to be a ZfitSpace, not {limits}."
        raise TypeError(msg)
    if counts is not None:
        if density is not None:
            msg = "Either specify 'counts' or 'density' but not both."
            raise ValueError(msg)
        is_density = False
        values = counts
    elif density is not None:
        is_density = True
        values = density
    else:
        msg = "Need to specify either 'counts' or 'density', not None."
        raise ValueError(msg)
    ndims = z._get_ndims(values)
    # partial = axis is not None and len(axis) < ndims
    if axis is not None:
        axis = convert_to_container(axis)
        if len(axis) > ndims:
            msg = f"axis {axis} is larger than values has ndims {values.shape}."
            raise ValueError(msg)
    else:
        axis = list(range(ndims))

    scaled_edges, (lower_bins, upper_bins), unscaled_edges = cut_edges_and_bins(
        edges=edges, limits=limits, axis=axis, unscaled=True
    )

    values_cut = tf.slice(values, lower_bins, (upper_bins - lower_bins))  # since limits are inclusive

    rank = values.shape.rank
    binwidths = []
    if not is_density:
        binwidths_unscaled = []
    # calculate the binwidth in each dimension
    for i, edge in enumerate(scaled_edges):
        edge_lower_index = [0] * rank
        # int32 is needed! Otherwise the gradient will fail
        edge_lowest_index = znp.array(edge_lower_index, dtype=znp.int32)

        edge_lower_index[i] = 1
        edge_lower_index = znp.array(edge_lower_index, dtype=znp.int32)
        edge_upper_index = [1] * rank
        edge_highest_index = edge_upper_index.copy()
        len_edge = tf.shape(edge)[i]
        edge_highest_index[i] = len_edge
        edge_highest_index = znp.asarray(edge_highest_index, dtype=znp.int32)
        edge_upper_index[i] = len_edge - 1  # len n -> index max is n - 1

        edge_upper_index = znp.asarray(edge_upper_index, dtype=znp.int32)
        lower_edge = tf.slice(edge, edge_lowest_index, (edge_upper_index - edge_lowest_index))
        upper_edge = tf.slice(edge, edge_lower_index, (edge_highest_index - edge_lower_index))
        binwidths.append(upper_edge - lower_edge)

        if not is_density:
            # unscaled edges to get the ratio
            lower_edge_unscaled = tf.slice(
                unscaled_edges[i],
                edge_lowest_index,
                (edge_upper_index - edge_lowest_index),
            )
            upper_edge_unscaled = tf.slice(
                unscaled_edges[i],
                edge_lower_index,
                (edge_highest_index - edge_lower_index),
            )
            binwidths_unscaled.append(upper_edge_unscaled - lower_edge_unscaled)

    binareas = reduce(operator.mul, binwidths)  # needs to be python as znp or tf can't broadcast otherwise
    if not is_density:  # scale the counts by the fraction. This is mostly one.
        binareas_uncut = reduce(operator.mul, binwidths_unscaled)
        # binareas_uncut = znp.prod(binwidths_unscaled, axis=0)
        binareas /= binareas_uncut
    values_cut *= binareas
    return tf.reduce_sum(values_cut, axis=axis)


@z.function(wraps="tensor", keepalive=True)
def cut_edges_and_bins(
    edges: Iterable[znp.array], limits: ZfitSpace, axis=None, unscaled=None
) -> tuple[list[znp.array], tuple[znp.array, znp.array], list | None]:
    """Cut the *edges* according to *limits* and calculate the bins inside.

    The edges within limits are calculated and returned together with the corresponding bin indices. The indices
    mark the lowest and the highest index of the edges that are returned. Additionally, the unscaled edges are returned.

    If the limits are between two edges, this will be treated as the new edge. If the limits are outside the edges,
    all edges in this direction will be returned (but not extended to the limit). For example:

    [0, 0.5, 1., 1.5, 2.] and the limits (0.8, 3.) will return [0.8, 1., 1.5, 2.], ([1], [4])

    .. code-block::

        cut_edges_and_bins([[0., 0.5, 1., 1.5, 2.]], ([[0.8]], [[3]]))



    Args:
        edges: Iterable of tensor-like objects that describe the edges of a histogram. Every object should have rank n
            (where n is the length of *edges*) but only have the dimension i filled out. These are
            tensors that are ready to be broadcasted together.
        limits: The limits that will be used to confine the edges


    Returns:
        edges, (lower bins, upper bins), unscaled_edges:  The edges and the bins are returned.
            The upper bin number corresponds to
            the highest bin which was still (partially) inside the limits **plus one** (so it's the index of the
            edge that is right outside). The unscaled edges are like *edges* but the last edge is the edge
            that is lying not inside anymore, so the actual edge of the last bin number returend.
            This can be used to determine the fraction cut away.
    """
    if axis is not None:
        axis = convert_to_container(axis)
    if unscaled is None:
        unscaled = False
    cut_unscaled_edges = [] if unscaled else None
    cut_scaled_edges = []

    all_lower_bins = []
    all_upper_bins = []
    if isinstance(limits, ZfitSpace):  # todo: check shapes?
        lower, upper = limits.v0.limits
    else:
        lower, upper = limits
        lower = znp.asarray(lower)
        upper = znp.asarray(upper)
    lower_all = lower[0]
    upper_all = upper[0]
    rank = len(edges)
    current_axis = 0
    for i, edge in enumerate(edges):
        edge = znp.asarray(edge)
        edge = znp.reshape(edge, (-1,))
        if axis is None or i in axis:
            lower_i = lower_all[current_axis, None]
            edge_minimum = edge[0]
            # edge_minimum = tf.gather(edge, indices=0, axis=i)
            lower_i = znp.maximum(lower_i, edge_minimum)
            upper_i = upper_all[current_axis, None]
            edge_maximum = edge[-1]
            # edge_maximum = tf.gather(edge, indices=tf.shape(edge)[i] - 1, axis=i)
            upper_i = znp.minimum(upper_i, edge_maximum)
            # we get the bins that are just one too far. Then we update this whole bin tensor with the actual edge.
            # The bins index is the index below the value.
            lower_bin_float = tfp.stats.find_bins(lower_i, edge, extend_lower_interval=True, extend_upper_interval=True)
            lower_bin = znp.reshape(znp.asarray(lower_bin_float, dtype=znp.int32), [-1])
            # lower_bins = tf.tensor_scatter_nd_update(zero_bins, [[i]], lower_bin)
            # +1 below because the outer bin is searched, meaning the one that is higher than the value

            upper_bin_float = tfp.stats.find_bins(upper_i, edge, extend_lower_interval=True, extend_upper_interval=True)
            upper_bin = znp.reshape(znp.asarray(upper_bin_float, dtype=znp.int32), [-1]) + 1
            size = upper_bin - lower_bin
            new_edge = tf.slice(edge, lower_bin, size + 1)  # +1 because stop is exclusive
            new_edge = tf.tensor_scatter_nd_update(new_edge, [tf.constant([0]), size], [lower_i[0], upper_i[0]])

            if unscaled:
                new_edge_unscaled = tf.slice(edge, lower_bin, size + 1)  # +1 because stop is exclusive

            current_axis += 1
        else:
            lower_bin = [0]
            upper_bin = znp.asarray([edge.shape[0] - 1], dtype=znp.int32)
            new_edge = edge
            if unscaled:
                new_edge_unscaled = edge
        new_shape = [1] * rank
        new_shape[i] = -1
        new_edge = znp.reshape(new_edge, new_shape)
        all_lower_bins.append(lower_bin)
        all_upper_bins.append(upper_bin)
        cut_scaled_edges.append(new_edge)
        if unscaled:
            new_edge_unscaled = znp.reshape(new_edge_unscaled, new_shape)
            cut_unscaled_edges.append(new_edge_unscaled)

    # partial = axis is not None and len(axis) < rank
    #
    # if partial:
    #     scaled_edges_full = list(edges)
    #     for edge, ax in zip(cut_scaled_edges, axis):
    #         scaled_edges_full[ax] = edge
    #     scaled_edges = scaled_edges_full
    #     indices = tf.convert_to_tensor(axis)[:, None]
    #     lower_bins = tf.scatter_nd(indices, lower_bins, shape=(ndims,))
    #     upper_bins = tf.tensor_scatter_nd_update(tf.convert_to_tensor(values.shape),
    #                                              indices, upper_bins)
    # lower_bins_indices = tf.stack([lower_bins, dims], axis=-1)
    # upper_bins_indices = tf.stack([upper_bins, dims], axis=-1)
    # all_lower_bins = znp.asarray(znp.sum(all_lower_bins, axis=0), dtype=znp.int32)
    all_lower_bins = tf.concat(all_lower_bins, axis=0)
    all_upper_bins = tf.concat(all_upper_bins, axis=0)
    return cut_scaled_edges, (all_lower_bins, all_upper_bins), cut_unscaled_edges



"""A rich selection of analytically implemented Distributions (models) are available in `TensorFlow Probability.

<https://github.com/tensorflow/probability>`_. While their API is slightly different from the zfit models, it is similar
enough to be easily wrapped.

Therefore, a convenient wrapper as well as a lot of implementations are provided.
"""

#  Copyright (c) 2024 zfit
from __future__ import annotations

from typing import Literal

import tensorflow as tf
import tensorflow_probability as tfp
import tensorflow_probability.python.distributions as tfd
from pydantic.v1 import Field

import zfit.z.numpy as znp
from zfit import z
from zfit.util.exception import AnalyticSamplingNotImplemented

from ..core.basepdf import BasePDF
from ..core.interfaces import ZfitData
from ..core.parameter import convert_to_parameter
from ..core.serialmixin import SerializableMixin
from ..core.space import Space, supports
from ..serialization import Serializer, SpaceRepr
from ..serialization.pdfrepr import BasePDFRepr
from ..settings import ztypes
from ..util import ztyping
from ..util.deprecation import deprecated_args
from ..util.ztyping import ExtendedInputType, NormInputType

# TODO: improve? while loop over `.sample`? Maybe as a fallback if not implemented?


def tfd_analytic_sample(n: int, dist: tfd.Distribution, limits: ztyping.ObsTypeInput):
    """Sample analytically with a `tfd.Distribution` within the limits. No preprocessing.

    Args:
        n: Number of samples to get
        dist: Distribution to sample from
        limits: Limits to sample from within

    Returns:
        The sampled data with the number of samples and the number of observables.
    """
    lower_bound, upper_bound = limits.v0.limits  # not working with MultiSpace
    lower_prob_lim = dist.cdf(lower_bound)
    upper_prob_lim = dist.cdf(upper_bound)

    shape = (n, 1)
    prob_sample = z.random.uniform(shape=shape, minval=lower_prob_lim, maxval=upper_prob_lim)
    prob_sample.set_shape((None, 1))
    try:
        sample = dist.quantile(prob_sample)
    except NotImplementedError:
        raise AnalyticSamplingNotImplemented from None
    sample.set_shape((None, limits.n_obs))
    return sample


class WrapDistribution(BasePDF):  # TODO: extend functionality of wrapper, like icdf
    """Baseclass to wrap tensorflow-probability distributions automatically."""

    def __init__(
        self,
        distribution,
        dist_params,
        obs,
        params=None,
        dist_kwargs=None,
        dtype=ztypes.float,
        name=None,
        **kwargs,
    ):
        # Check if subclass of distribution?
        if dist_kwargs is None:
            dist_kwargs = {}

        if dist_params is None:
            dist_params = {}
        name = name or distribution.name
        params = dist_params.copy() if params is None else {k: convert_to_parameter(p) for k, p in params.items()}

        super().__init__(obs=obs, dtype=dtype, name=name, params=params, **kwargs)

        self._distribution = distribution
        self.dist_params = dist_params
        self.dist_kwargs = dist_kwargs
        self._inverse_analytic_integral = []

    @property
    def distribution(self):
        params = self.dist_params
        if callable(params):
            params = params()
        kwargs = self.dist_kwargs
        if callable(kwargs):
            kwargs = kwargs()
        return self._distribution(**params, **kwargs, name=self.name + "_tfp")

    def _unnormalized_pdf(self, x: ZfitData):
        value = z.unstack_x(x)  # TODO: use this? change shaping below?
        return self.distribution.prob(value=value, name="unnormalized_pdf")

    # TODO: register integral?
    @supports()
    def _analytic_integrate(self, limits, norm):
        del norm  # not supported
        lower, upper = limits._rect_limits_tf
        lower = z.unstack_x(lower)
        upper = z.unstack_x(upper)
        tf.debugging.assert_all_finite((lower, upper), "Are infinite limits needed? Causes troubles with NaNs")
        return self.distribution.cdf(upper) - self.distribution.cdf(lower)

    def _analytic_sample(self, n, limits: Space):
        return tfd_analytic_sample(n=n, dist=self.distribution, limits=limits)


# class KernelDensityTFP(WrapDistribution):
#
#     def __init__(self, loc: ztyping.ParamTypeInput, scale: ztyping.ParamTypeInput, obs: ztyping.ObsTypeInput,
#                  kernel: tfp.distributions.Distribution = tfp.distributions.Normal,
#                  weights: Union[None, np.ndarray, tf.Tensor] = None, name: str = "KernelDensity"):
#         """Kernel Density Estimation of loc and either a broadcasted or a per-loc scale with a Distribution as kernel.
#
#         Args:
#             loc: 1-D Tensor-like. The positions of the `kernel`. Determines how many kernels will be created.
#             scale: Broadcastable to the batch and event shape of the distribution. A scalar will simply broadcast
#                 to `loc` for a 1-D distribution.
#             obs: Observables
#             kernel: Distribution that is used as kernel
#             weights: Weights of each `loc`, can be None or Tensor-like with shape compatible with loc
#             name: Name of the PDF
#         """
#         if not isinstance(kernel,
#                           tfp.distributions.Distribution) and False:  # HACK remove False, why does test not work?
#             raise TypeError("Currently, only tfp distributions are supported as kernels. Please open an issue if this "
#                             "is too restrictive.")
#
#         if isinstance(loc, ZfitData):
#             if loc.weights is not None:
#                 if weights is not None:
#                     raise OverdefinedError("Cannot specify weights and use a `ZfitData` with weights.")
#                 else:
#                     weights = loc.weights
#
#         if weights is None:
#             weights = tf.ones_like(loc, dtype=tf.float64)
#         self._weights_loc = weights
#         self._weights_sum = z.reduce_sum(weights)
#         self._latent_loc = loc
#         params = {"scale": scale}
#         dist_params = {"loc": loc, "scale": scale}
#         super().__init__(distribution=kernel, dist_params=dist_params, obs=obs, params=params, dtype=ztypes.float,
#                          name=name)
#
#     def _unnormalized_pdf(self, x: "zfit.Data", norm_range=False):
#         value = znp.expand_dims(x.value(), -2)
#         new_shape = znp.concatenate([tf.shape(value)[:2], [tf.shape(self._latent_loc)[0], 4]], axis=0)
#         value = tf.broadcast_to(value, new_shape)
#         probs = self.distribution.prob(value=value, name="unnormalized_pdf")
#         # weights = znp.expand_dims(self._weights_loc, axis=-1)
#         weights = self._weights_loc
#         probs = z.reduce_sum(probs * weights, axis=-1) / self._weights_sum
#         return probs
#
#     @supports()
#     def _analytic_integrate(self, limits, norm_range):
#         lower, upper = limits.limits
#         if np.all(-np.array(lower) == np.array(upper)) and np.all(np.array(upper) == np.infty):
#             return z.reduce_sum(self._weights_loc)  # tfp distributions are normalized to 1
#         lower = z.to_real(lower[0], dtype=self.dtype)
#         # lower = tf.broadcast_to(lower, shape=(tf.shape(self._latent_loc)[0], limits.n_obs,))  # remove
#         upper = z.to_real(upper[0], dtype=self.dtype)
#         integral = self.distribution.cdf(upper) - self.distribution.cdf(lower)
#         integral = z.reduce_sum(integral * self._weights_loc, axis=-1) / self._weights_sum
#         return integral  # TODO: generalize for VectorSpaces


class Gauss(WrapDistribution, SerializableMixin):
    _N_OBS = 1

    def __init__(
        self,
        mu: ztyping.ParamTypeInput,
        sigma: ztyping.ParamTypeInput,
        obs: ztyping.ObsTypeInput,
        *,
        extended: ExtendedInputType = None,
        norm: NormInputType = None,
        name: str = "Gauss",
        label=None,
    ):
        """Gaussian or Normal distribution with a mean (mu) and a standard deviation (sigma).

        The gaussian shape is defined as

        .. math::
            f(x \\mid \\mu, \\sigma^2) = e^{ -\\frac{(x - \\mu)^{2}}{2\\sigma^2} }

        with the normalization over [-inf, inf] of

        .. math::
            \\frac{1}{\\sqrt{2\\pi\\sigma^2} }

        The normalization changes for different normalization ranges

        Args:
            mu: Mean of the gaussian dist
            sigma: Standard deviation or spread of the gaussian
            obs: |@doc:pdf.init.obs| Observables of the
               model. This will be used as the default space of the PDF and,
               if not given explicitly, as the normalization range.

               The default space is used for example in the sample method: if no
               sampling limits are given, the default space is used.

               If the observables are binned and the model is unbinned, the
               model will be a binned model, by wrapping the model in a
               :py:class:`~zfit.pdf.BinnedFromUnbinnedPDF`, equivalent to
               calling :py:meth:`~zfit.pdf.BasePDF.to_binned`.

               If the observables are binned and the model is unbinned, the
               model will be a binned model, by wrapping the model in a
               :py:class:`~zfit.pdf.BinnedFromUnbinnedPDF`, equivalent to
               calling :py:meth:`~zfit.pdf.BasePDF.to_binned`.

               The observables are not equal to the domain as it does not restrict or
               truncate the model outside this range. |@docend:pdf.init.obs|
            extended: |@doc:pdf.init.extended| The overall yield of the PDF.
               If this is parameter-like, it will be used as the yield,
               the expected number of events, and the PDF will be extended.
               An extended PDF has additional functionality, such as the
               ``ext_*`` methods and the ``counts`` (for binned PDFs). |@docend:pdf.init.extended|
            norm: |@doc:pdf.init.norm| Normalization of the PDF.
               By default, this is the same as the default space of the PDF. |@docend:pdf.init.norm|
            name: |@doc:pdf.init.name| Name of the PDF.
               Maybe has implications on the serialization and deserialization of the PDF.
               For a human-readable name, use the label. |@docend:pdf.init.name|
            label: |@doc:pdf.init.label| Human-readable name
               or label of
               the PDF for a better description, to be used with plots etc.
               Has no programmatical functional purpose as identification. |@docend:pdf.init.label|
        """
        mu, sigma = self._check_input_params_tfp(mu, sigma)
        params = {"mu": mu, "sigma": sigma}

        def dist_params():
            return {"loc": mu.value(), "scale": sigma.value()}

        distribution = tfp.distributions.Normal
        super().__init__(
            distribution=distribution,
            dist_params=dist_params,
            obs=obs,
            params=params,
            name=name,
            extended=extended,
            norm=norm,
            label=label,
        )


class GaussPDFRepr(BasePDFRepr):
    _implementation = Gauss
    hs3_type: Literal["Gauss"] = Field("Gauss", alias="type")
    x: SpaceRepr
    mu: Serializer.types.ParamInputTypeDiscriminated
    sigma: Serializer.types.ParamInputTypeDiscriminated


class ExponentialTFP(WrapDistribution):
    _N_OBS = 1

    def __init__(
        self,
        tau: ztyping.ParamTypeInput,
        obs: ztyping.ObsTypeInput,
        name: str = "Exponential",
    ):
        (tau,) = self._check_input_params_tfp(tau)
        params = {"tau", tau}
        dist_params = {"rate": tau}
        distribution = tfp.distributions.Exponential
        super().__init__(
            distribution=distribution,
            dist_params=dist_params,
            obs=obs,
            params=params,
            name=name,
        )


class Uniform(WrapDistribution):
    _N_OBS = 1

    def __init__(
        self,
        low: ztyping.ParamTypeInput,
        high: ztyping.ParamTypeInput,
        obs: ztyping.ObsTypeInput,
        *,
        extended: ExtendedInputType = None,
        norm: NormInputType = None,
        name: str = "Uniform",
        label: str | None = None,
    ):
        """Uniform distribution which is constant between `low`, `high` and zero outside.

        Args:
            low: Below this value, the pdf is zero.
            high: Above this value, the pdf is zero.
            obs: Observables and normalization range the pdf is defined in
            extended: |@doc:pdf.init.extended| The overall yield of the PDF.
               If this is parameter-like, it will be used as the yield,
               the expected number of events, and the PDF will be extended.
               An extended PDF has additional functionality, such as the
               ``ext_*`` methods and the ``counts`` (for binned PDFs). |@docend:pdf.init.extended|
            norm: |@doc:pdf.init.norm| Normalization of the PDF.
               By default, this is the same as the default space of the PDF. |@docend:pdf.init.norm|
            name: |@doc:pdf.init.name| Name of the PDF.
               Maybe has implications on the serialization and deserialization of the PDF.
               For a human-readable name, use the label. |@docend:pdf.init.name|
            label: |@doc:pdf.init.label| Human-readable name
               or label of
               the PDF for a better description, to be used with plots etc.
               Has no programmatical functional purpose as identification. |@docend:pdf.init.label|
        """
        low, high = self._check_input_params_tfp(low, high)
        params = {"low": low, "high": high}

        def dist_params():
            return {"low": low.value(), "high": high.value()}

        distribution = tfp.distributions.Uniform
        super().__init__(
            distribution=distribution,
            dist_params=dist_params,
            obs=obs,
            params=params,
            name=name,
            extended=extended,
            norm=norm,
            label=label,
        )


class TruncatedGauss(WrapDistribution):
    _N_OBS = 1

    def __init__(
        self,
        mu: ztyping.ParamTypeInput,
        sigma: ztyping.ParamTypeInput,
        low: ztyping.ParamTypeInput,
        high: ztyping.ParamTypeInput,
        obs: ztyping.ObsTypeInput,
        *,
        extended: ExtendedInputType = None,
        norm: NormInputType = None,
        name: str = "TruncatedGauss",
        label: str | None = None,
    ):
        """Gaussian distribution that is 0 outside of `low`, `high`. Equivalent to the product of Gauss and Uniform.

        Args:
            mu: Mean of the gaussian dist
            sigma: Standard deviation or spread of the gaussian
            low: Below this value, the pdf is zero.
            high: Above this value, the pdf is zero.
            obs: Observables and normalization range the pdf is defined in
            extended: |@doc:pdf.init.extended| The overall yield of the PDF.
               If this is parameter-like, it will be used as the yield,
               the expected number of events, and the PDF will be extended.
               An extended PDF has additional functionality, such as the
               ``ext_*`` methods and the ``counts`` (for binned PDFs). |@docend:pdf.init.extended|
            norm: |@doc:pdf.init.norm| Normalization of the PDF.
               By default, this is the same as the default space of the PDF. |@docend:pdf.init.norm|
            name: |@doc:pdf.init.name| Name of the PDF.
               Maybe has implications on the serialization and deserialization of the PDF.
               For a human-readable name, use the label. |@docend:pdf.init.name|
            label: |@doc:pdf.init.label| Human-readable name
               or label of
               the PDF for a better description, to be used with plots etc.
               Has no programmatical functional purpose as identification. |@docend:pdf.init.label|
        """
        mu, sigma, low, high = self._check_input_params_tfp(mu, sigma, low, high)
        params = {"mu": mu, "sigma": sigma, "low": low, "high": high}
        distribution = tfp.distributions.TruncatedNormal

        def dist_params():
            return {
                "loc": mu.value(),
                "scale": sigma.value(),
                "low": low.value(),
                "high": high.value(),
            }

        super().__init__(
            distribution=distribution,
            dist_params=dist_params,
            obs=obs,
            params=params,
            name=name,
            extended=extended,
            norm=norm,
            label=label,
        )


class Cauchy(WrapDistribution, SerializableMixin):
    _N_OBS = 1

    def __init__(
        self,
        m: ztyping.ParamTypeInput,
        gamma: ztyping.ParamTypeInput,
        obs: ztyping.ObsTypeInput,
        *,
        extended: ExtendedInputType = None,
        norm: NormInputType = None,
        name: str = "Cauchy",
        label: str | None = None,
    ):
        r"""Non-relativistic Breit-Wigner (Cauchy) PDF representing the energy distribution of a decaying particle.

        The (unnormalized) shape of the non-relativistic Breit-Wigner is given by

        .. math::

            \frac{1}{\gamma \left[1 + \left(\frac{x - m}{\gamma}\right)^2\right]}

        with :math:`m` the mean and :math:`\gamma` the width of the distribution.

        Args:
            m: Invariant mass of the unstable particle.
            gamma: Width of the shape.
            obs: Observables and normalization range the pdf is defined in
            extended: |@doc:pdf.init.extended| The overall yield of the PDF.
               If this is parameter-like, it will be used as the yield,
               the expected number of events, and the PDF will be extended.
               An extended PDF has additional functionality, such as the
               ``ext_*`` methods and the ``counts`` (for binned PDFs). |@docend:pdf.init.extended|
            norm: |@doc:pdf.init.norm| Normalization of the PDF.
               By default, this is the same as the default space of the PDF. |@docend:pdf.init.norm|
            name: |@doc:pdf.init.name| Name of the PDF.
               Maybe has implications on the serialization and deserialization of the PDF.
               For a human-readable name, use the label. |@docend:pdf.init.name|
            label: |@doc:pdf.init.label| Human-readable name
               or label of
               the PDF for a better description, to be used with plots etc.
               Has no programmatical functional purpose as identification. |@docend:pdf.init.label|
        """
        m, gamma = self._check_input_params_tfp(m, gamma)
        params = {"m": m, "gamma": gamma}
        distribution = tfp.distributions.Cauchy

        def dist_params():
            return {"loc": m.value(), "scale": gamma.value()}

        super().__init__(
            distribution=distribution,
            dist_params=dist_params,
            obs=obs,
            params=params,
            name=name,
            extended=extended,
            norm=norm,
            label=label,
        )


class CauchyPDFRepr(BasePDFRepr):
    _implementation = Cauchy
    hs3_type: Literal["Cauchy"] = Field("Cauchy", alias="type")
    x: SpaceRepr
    m: Serializer.types.ParamTypeDiscriminated
    gamma: Serializer.types.ParamTypeDiscriminated


class Poisson(WrapDistribution, SerializableMixin):
    _N_OBS = 1

    @deprecated_args(None, "Use lam instead", "lamb")
    def __init__(
        self,
        lam: ztyping.ParamTypeInput = None,
        obs: ztyping.ObsTypeInput = None,
        *,
        extended: ExtendedInputType = None,
        norm: NormInputType = None,
        name: str = "Poisson",
        lamb=None,
        label: str | None = None,
    ):
        """Poisson distribution, parametrized with an event rate parameter (lamb).

        The probability mass function of the Poisson distribution is given by

        .. math::
            f(x, \\lambda) = \\frac{\\lambda^{x}e^{-\\lambda}}{x!}

        Args:
            lamb: the event rate
            obs: Observables and normalization range the pdf is defined in
            extended: |@doc:pdf.init.extended| The overall yield of the PDF.
               If this is parameter-like, it will be used as the yield,
               the expected number of events, and the PDF will be extended.
               An extended PDF has additional functionality, such as the
               ``ext_*`` methods and the ``counts`` (for binned PDFs). |@docend:pdf.init.extended|
            norm: |@doc:pdf.init.norm| Normalization of the PDF.
               By default, this is the same as the default space of the PDF. |@docend:pdf.init.norm|
            name: |@doc:pdf.init.name| Name of the PDF.
               Maybe has implications on the serialization and deserialization of the PDF.
               For a human-readable name, use the label. |@docend:pdf.init.name|
            label: |@doc:pdf.init.label| Human-readable name
               or label of
               the PDF for a better description, to be used with plots etc.
               Has no programmatical functional purpose as identification. |@docend:pdf.init.label|
        """
        if lamb is not None:
            lam = lamb
        del lamb
        (lam,) = self._check_input_params_tfp(lam)
        params = {"lam": lam}

        def dist_params():
            return {"rate": lam.value()}

        distribution = tfp.distributions.Poisson
        super().__init__(
            distribution=distribution,
            dist_params=dist_params,
            obs=obs,
            params=params,
            name=name,
            extended=extended,
            norm=norm,
            label=label,
        )


class PoissonPDFRepr(BasePDFRepr):
    _implementation = Poisson
    hs3_type: Literal["Poisson"] = Field("Poisson", alias="type")
    x: SpaceRepr
    lam: Serializer.types.ParamTypeDiscriminated


class LogNormal(WrapDistribution, SerializableMixin):
    _N_OBS = 1

    def __init__(
        self,
        mu: ztyping.ParamTypeInput,
        sigma: ztyping.ParamTypeInput,
        obs: ztyping.ObsTypeInput,
        *,
        extended: ExtendedInputType = None,
        norm: NormInputType = None,
        name: str = "LogNormal",
        label: str | None = None,
    ):
        r"""Log-normal distribution, the exponential of a normal distribution.

        The probability density function of the log-normal distribution is only defined for positive values and
        is given by

        .. math::
            f(x \\mid \mu, \sigma) = \frac{1}{x \sigma \sqrt{2\pi}} e^{-\frac{(\ln(x) - \mu)^2}{2\sigma^2}}

        with :math:`\mu` the mean and :math:`\sigma` the standard deviation of the underlying normal distribution.

        Args:
            mu: Mean of the underlying normal distribution.
            sigma: Standard deviation of the underlying normal distribution.
            obs: Observables and normalization range the pdf is defined in
            extended: |@doc:pdf.init.extended| The overall yield of the PDF.
               If this is parameter-like, it will be used as the yield,
               the expected number of events, and the PDF will be extended.
               An extended PDF has additional functionality, such as the
               ``ext_*`` methods and the ``counts`` (for binned PDFs). |@docend:pdf.init.extended|
            norm: |@doc:pdf.init.norm| Normalization of the PDF.
               By default, this is the same as the default space of the PDF. |@docend:pdf.init.norm|
            name: |@doc:pdf.init.name| Name of the PDF.
               Maybe has implications on the serialization and deserialization of the PDF.
               For a human-readable name, use the label. |@docend:pdf.init.name|
            label: |@doc:pdf.init.label| Human-readable name
               or label of
               the PDF for a better description, to be used with plots etc.
               Has no programmatical functional purpose as identification. |@docend:pdf.init.label|
        """
        mu, sigma = self._check_input_params_tfp(mu, sigma)

        params = {"mu": mu, "sigma": sigma}

        def dist_params():
            return {"loc": mu.value(), "scale": sigma.value()}

        distribution = tfp.distributions.LogNormal
        super().__init__(
            distribution=distribution,
            dist_params=dist_params,
            obs=obs,
            params=params,
            name=name,
            extended=extended,
            norm=norm,
            label=label,
        )


class LogNormalPDFRepr(BasePDFRepr):
    _implementation = LogNormal
    hs3_type: Literal["LogNormal"] = Field("LogNormal", alias="type")
    x: SpaceRepr
    mu: Serializer.types.ParamTypeDiscriminated
    sigma: Serializer.types.ParamTypeDiscriminated


class ChiSquared(WrapDistribution, SerializableMixin):
    _N_OBS = 1

    def __init__(
        self,
        ndof: ztyping.ParamTypeInput,
        obs: ztyping.ObsTypeInput,
        *,
        extended: ExtendedInputType = None,
        norm: NormInputType = None,
        name: str = "ChiSquared",
        label: str | None = None,
    ):
        """ChiSquared distribution for ndof degrees of freedom.

        The chisquared shape for `d` degrees of freedom is defined as

        .. math::

            f(x \\mid d) = x^(d/2 - 1) \\exp(-x/2) / Z

        with the normalization over [0, inf] of

        .. math::

            Z = \\frac{1}{2^{d/2} \\Gamma(d/2)}

        The normalization changes for different normalization ranges

        Args:
            ndof: Number of degrees of freedom
            obs: Observables and normalization range the pdf is defined in
            extended: |@doc:pdf.init.extended| The overall yield of the PDF.
               If this is parameter-like, it will be used as the yield,
               the expected number of events, and the PDF will be extended.
               An extended PDF has additional functionality, such as the
               ``ext_*`` methods and the ``counts`` (for binned PDFs). |@docend:pdf.init.extended|
            norm: |@doc:pdf.init.norm| Normalization of the PDF.
               By default, this is the same as the default space of the PDF. |@docend:pdf.init.norm|
            name: |@doc:pdf.init.name| Name of the PDF.
               Maybe has implications on the serialization and deserialization of the PDF.
               For a human-readable name, use the label. |@docend:pdf.init.name|
            label: |@doc:pdf.init.label| Human-readable name
               or label of
               the PDF for a better description, to be used with plots etc.
               Has no programmatical functional purpose as identification. |@docend:pdf.init.label|
        """
        (ndof,) = self._check_input_params_tfp(ndof)
        params = {"ndof": ndof}

        def dist_params():
            return {"df": ndof.value()}

        distribution = tfp.distributions.Chi2
        super().__init__(
            distribution=distribution,
            dist_params=dist_params,
            obs=obs,
            params=params,
            name=name,
            extended=extended,
            norm=norm,
            label=label,
        )


class ChiSquaredPDFRepr(BasePDFRepr):
    _implementation = ChiSquared
    hs3_type: Literal["ChiSquared"] = Field("ChiSquared", alias="type")
    x: SpaceRepr
    ndof: Serializer.types.ParamTypeDiscriminated


class StudentT(WrapDistribution, SerializableMixin):
    _N_OBS = 1

    def __init__(
        self,
        ndof: ztyping.ParamTypeInput,
        mu: ztyping.ParamTypeInput,
        sigma: ztyping.ParamTypeInput,
        obs: ztyping.ObsTypeInput,
        extended: ExtendedInputType = None,
        norm: NormInputType = None,
        name: str = "StudentT",
        label: str | None = None,
    ):
        """StudentT distribution for ndof degrees of freedom.

        The StudentT shape for `d` degrees of freedom is defined as

        .. math::

            f(x \\mid d, \\mu, \\sigma) = \\left(1 + \\frac{1}{d} \\left(\\frac{x - \\mu}{\\sigma}\\right)^2\\right)^{-\\frac{d+1}{2}} / Z

        with the normalization over [-inf, inf] of

        .. math::

            Z = \\frac{\\sqrt{d \\pi} \\Gamma(\\frac{d}{2})}{\\Gamma(\\frac{d+1}{2})}

        The normalization changes for different normalization ranges

        Args:
            ndof: Number of degrees of freedom
            mu: Mean of the distribution
            sigma: Scale of the distribution
            obs: |@doc:pdf.init.obs| Observables of the
               model. This will be used as the default space of the PDF and,
               if not given explicitly, as the normalization range.

               The default space is used for example in the sample method: if no
               sampling limits are given, the default space is used.

               If the observables are binned and the model is unbinned, the
               model will be a binned model, by wrapping the model in a
               :py:class:`~zfit.pdf.BinnedFromUnbinnedPDF`, equivalent to
               calling :py:meth:`~zfit.pdf.BasePDF.to_binned`.

               If the observables are binned and the model is unbinned, the
               model will be a binned model, by wrapping the model in a
               :py:class:`~zfit.pdf.BinnedFromUnbinnedPDF`, equivalent to
               calling :py:meth:`~zfit.pdf.BasePDF.to_binned`.

               The observables are not equal to the domain as it does not restrict or
               truncate the model outside this range. |@docend:pdf.init.obs|
            extended: |@doc:pdf.init.extended| The overall yield of the PDF.
               If this is parameter-like, it will be used as the yield,
               the expected number of events, and the PDF will be extended.
               An extended PDF has additional functionality, such as the
               ``ext_*`` methods and the ``counts`` (for binned PDFs). |@docend:pdf.init.extended|
            norm: |@doc:pdf.init.norm| Normalization of the PDF.
               By default, this is the same as the default space of the PDF. |@docend:pdf.init.norm|
            name: |@doc:pdf.init.name| Name of the PDF.
               Maybe has implications on the serialization and deserialization of the PDF.
               For a human-readable name, use the label. |@docend:pdf.init.name|
            label: |@doc:pdf.init.label| Human-readable name
               or label of
               the PDF for a better description, to be used with plots etc.
               Has no programmatical functional purpose as identification. |@docend:pdf.init.label|
        """
        ndof, mu, sigma = self._check_input_params_tfp(ndof, mu, sigma)
        params = {"ndof": ndof, "mu": mu, "sigma": sigma}

        def dist_params():
            return {"df": ndof.value(), "loc": mu.value(), "scale": sigma.value()}

        distribution = tfp.distributions.StudentT
        super().__init__(
            distribution=distribution,
            dist_params=dist_params,
            obs=obs,
            params=params,
            name=name,
            extended=extended,
            norm=norm,
            label=label,
        )


class StudentTPDFRepr(BasePDFRepr):
    _implementation = StudentT
    hs3_type: Literal["StudentT"] = Field("StudentT", alias="type")
    x: SpaceRepr
    ndof: Serializer.types.ParamTypeDiscriminated
    mu: Serializer.types.ParamTypeDiscriminated
    sigma: Serializer.types.ParamTypeDiscriminated


class QGauss(WrapDistribution, SerializableMixin):
    _N_OBS = 1

    def __init__(
        self,
        q: ztyping.ParamTypeInput,
        mu: ztyping.ParamTypeInput,
        sigma: ztyping.ParamTypeInput,
        obs: ztyping.ObsTypeInput,
        *,
        extended: ExtendedInputType = None,
        norm: NormInputType = None,
        name: str = "QGauss",
        label: str | None = None,
    ):
        """Q-Gaussian distribution with parameter `q`.

        The q-Gaussian is a probability distribution arising from the maximization of the Tsallis entropy under appropriate constraints.
        It is defined for q < 3 and the Gaussian distribution is recovered as q -> 1.
        For q < 1, is it the PDF of a bounded random variable.
        We only support 1 < q < 3 in this implementation.
        If you want to use exactly q = 1, use the `zfit.pdf.Gauss` class.
        During fitting, if you want to start from a Gaussian shape, you can initialize the `q` parameter to be really close to 1.
        It is related to the Student's t-distribution according to the `corresponding Wikipedia entry <https://en.wikipedia.org/wiki/Q-Gaussian_distribution#Student's_t-distribution>`_
        and that is how it is implemented here.

        The q-Gaussian shape for 1 < q < 3 is defined as

        .. math::

            f(x \\mid q, \\mu, \\sigma) = \\frac{1}{C_{q} \\sigma} e_{q}\\left(-\\left(\\frac{x - \\mu}{\\sigma}\\right)^{2}\\right)

        with

        .. math::

            e_q(x) = \\left[1 + (1 - q) x\\right]_{+}^{\\frac{1}{1 - q}}

        and the normalization over [-inf, inf] of

        .. math::

            C_{q} = \\frac{\\sqrt{\\pi} \\Gamma \\left(\\frac{3 - q}{2 (q - 1)}\\right)}{\\sqrt{q - 1}\\Gamma \\left(\\frac{1}{q - 1}\\right)}

        The normalization changes for different normalization ranges

        Args:
            q: Shape parameter of the q-Gaussian. Must be 1 < q < 3.
            mu: Mean of the distribution
            sigma: Scale of the distribution
            obs: |@doc:pdf.init.obs| Observables of the
               model. This will be used as the default space of the PDF and,
               if not given explicitly, as the normalization range.

               The default space is used for example in the sample method: if no
               sampling limits are given, the default space is used.

               If the observables are binned and the model is unbinned, the
               model will be a binned model, by wrapping the model in a
               :py:class:`~zfit.pdf.BinnedFromUnbinnedPDF`, equivalent to
               calling :py:meth:`~zfit.pdf.BasePDF.to_binned`.

               If the observables are binned and the model is unbinned, the
               model will be a binned model, by wrapping the model in a
               :py:class:`~zfit.pdf.BinnedFromUnbinnedPDF`, equivalent to
               calling :py:meth:`~zfit.pdf.BasePDF.to_binned`.

               The observables are not equal to the domain as it does not restrict or
               truncate the model outside this range. |@docend:pdf.init.obs|
            extended: |@doc:pdf.init.extended| The overall yield of the PDF.
               If this is parameter-like, it will be used as the yield,
               the expected number of events, and the PDF will be extended.
               An extended PDF has additional functionality, such as the
               ``ext_*`` methods and the ``counts`` (for binned PDFs). |@docend:pdf.init.extended|
            norm: |@doc:pdf.init.norm| Normalization of the PDF.
               By default, this is the same as the default space of the PDF. |@docend:pdf.init.norm|
            name: |@doc:pdf.init.name| Name of the PDF.
               Maybe has implications on the serialization and deserialization of the PDF.
               For a human-readable name, use the label. |@docend:pdf.init.name|
            label: |@doc:pdf.init.label| Human-readable name
               or label of
               the PDF for a better description, to be used with plots etc.
               Has no programmatical functional purpose as identification. |@docend:pdf.init.label|
        """
        from zfit import run

        q, mu, sigma = self._check_input_params_tfp(q, mu, sigma)
        if run.executing_eagerly():
            if q < 1 or q > 3:
                msg = "q < 1 or q > 3 are not supported"
                raise ValueError(msg)
            if q == 1:
                msg = "q = 1 is a Gaussian, use Gauss instead."
                raise ValueError(msg)
        elif run.numeric_checks:
            tf.debugging.assert_greater(q, znp.asarray(1.0), "q must be > 1")
            tf.debugging.assert_less(q, znp.asarray(3.0), "q must be < 3")
        params = {"q": q, "mu": mu, "sigma": sigma}

        # https://en.wikipedia.org/wiki/Q-Gaussian_distribution
        # relation to Student's t-distribution

        # 1/(2 sigma^2) = 1 / (3 - q)
        # 2 sigma^2 = 3 - q
        # sigma = sqrt((3 - q)/2)

        def dist_params(q=q, mu=mu, sigma=sigma):
            if run.numeric_checks:
                tf.debugging.assert_greater(q, znp.asarray(1.0), "q must be > 1")
                tf.debugging.assert_less(q, znp.asarray(3.0), "q must be < 3")
            df = (3 - q.value()) / (q.value() - 1)
            scale = sigma.value() / tf.sqrt(0.5 * (3 - q.value()))
            return {"df": df, "loc": mu.value(), "scale": scale}

        distribution = tfp.distributions.StudentT
        super().__init__(
            distribution=distribution,
            dist_params=dist_params,
            obs=obs,
            params=params,
            name=name,
            extended=extended,
            norm=norm,
            label=label,
        )


class QGaussPDFRepr(BasePDFRepr):
    _implementation = QGauss
    hs3_type: Literal["QGauss"] = Field("QGauss", alias="type")
    x: SpaceRepr
    q: Serializer.types.ParamTypeDiscriminated
    mu: Serializer.types.ParamTypeDiscriminated
    sigma: Serializer.types.ParamTypeDiscriminated


class BifurGauss(WrapDistribution, SerializableMixin):
    _N_OBS = 1

    def __init__(
        self,
        mu: ztyping.ParamTypeInput,
        sigmal: ztyping.ParamTypeInput,
        sigmar: ztyping.ParamTypeInput,
        obs: ztyping.ObsTypeInput,
        *,
        extended: ExtendedInputType = None,
        norm: NormInputType = None,
        name: str = "BifurGauss",
        label: str | None = None,
    ):
        """Bifurcated Gaussian distribution different standard deviations for the left and right side of the mean.

        The bifurcated Gaussian shape is defined as

        .. math::

            f(x \\mid \\mu, \\sigma_{L}, \\sigma_{R}) = \\begin{cases}
            A \\exp{\\left(-\\frac{(x - \\mu)^2}{2 \\sigma_{L}^2}\\right)}, & \\mbox{for } x < \\mu \\newline
            A \\exp{\\left(-\\frac{(x - \\mu)^2}{2 \\sigma_{R}^2}\\right)}, & \\mbox{for } x \\geq \\mu
            \\end{cases}

        with the normalization over [-inf, inf] of

        .. math::

            A = \\sqrt{\\frac{2}{\\pi}} \\frac{1}{\\sigma_{L} + \\sigma_{R}}

        The normalization changes for different normalization ranges

        Args:
            mu: Mean of the distribution
            sigmal: Standard deviation on the left side of the mean
            sigmar: Standard deviation for the right side of the mean
            obs: |@doc:pdf.init.obs| Observables of the
               model. This will be used as the default space of the PDF and,
               if not given explicitly, as the normalization range.

               The default space is used for example in the sample method: if no
               sampling limits are given, the default space is used.

               If the observables are binned and the model is unbinned, the
               model will be a binned model, by wrapping the model in a
               :py:class:`~zfit.pdf.BinnedFromUnbinnedPDF`, equivalent to
               calling :py:meth:`~zfit.pdf.BasePDF.to_binned`.

               If the observables are binned and the model is unbinned, the
               model will be a binned model, by wrapping the model in a
               :py:class:`~zfit.pdf.BinnedFromUnbinnedPDF`, equivalent to
               calling :py:meth:`~zfit.pdf.BasePDF.to_binned`.

               The observables are not equal to the domain as it does not restrict or
               truncate the model outside this range. |@docend:pdf.init.obs|
            extended: |@doc:pdf.init.extended| The overall yield of the PDF.
               If this is parameter-like, it will be used as the yield,
               the expected number of events, and the PDF will be extended.
               An extended PDF has additional functionality, such as the
               ``ext_*`` methods and the ``counts`` (for binned PDFs). |@docend:pdf.init.extended|
            norm: |@doc:pdf.init.norm| Normalization of the PDF.
               By default, this is the same as the default space of the PDF. |@docend:pdf.init.norm|
            name: |@doc:pdf.init.name| Name of the PDF.
               Maybe has implications on the serialization and deserialization of the PDF.
               For a human-readable name, use the label. |@docend:pdf.init.name|
            label: |@doc:pdf.init.label| Human-readable name
               or label of
               the PDF for a better description, to be used with plots etc.
               Has no programmatical functional purpose as identification. |@docend:pdf.init.label|
        """
        mu, sigmal, sigmar = self._check_input_params_tfp(mu, sigmal, sigmar)
        params = {"mu": mu, "sigmal": sigmal, "sigmar": sigmar}

        # sigmal = scale / skewness
        # sigmar = scale * skewness
        # scale = sigmal * skewness
        # sigmar = sigmal * skewness^2
        # skewness = sqrt(sigmar / sigmal)
        # scale = sigmal * sqrt(sigmar / sigmal)

        def dist_params():
            scale = sigmal.value() * znp.sqrt(sigmar.value() / sigmal.value())
            skewness = znp.sqrt(sigmar.value() / sigmal.value())
            return {"loc": mu.value(), "scale": scale, "skewness": skewness}

        distribution = tfp.distributions.TwoPieceNormal
        super().__init__(
            distribution=distribution,
            dist_params=dist_params,
            obs=obs,
            params=params,
            name=name,
            extended=extended,
            norm=norm,
            label=label,
        )


class BifurGaussPDFRepr(BasePDFRepr):
    _implementation = BifurGauss
    hs3_type: Literal["BifurGauss"] = Field("BifurGauss", alias="type")
    x: SpaceRepr
    mu: Serializer.types.ParamTypeDiscriminated
    sigmal: Serializer.types.ParamTypeDiscriminated
    sigmar: Serializer.types.ParamTypeDiscriminated


class Gamma(WrapDistribution, SerializableMixin):
    _N_OBS = 1

    def __init__(
        self,
        gamma: ztyping.ParamTypeInput,
        beta: ztyping.ParamTypeInput,
        mu: ztyping.ParamTypeInput,
        obs: ztyping.ObsTypeInput,
        *,
        extended: ExtendedInputType = None,
        norm: NormInputType = None,
        name: str = "Gamma",
        label: str | None = None,
    ):
        """Gamma distribution.

        The gamma shape is parametrized here with `gamma`, `beta` and `mu`, following
        the same parametrization `as RooFit <https://root.cern.ch/doc/master/classRooGamma.html>`_.
        The gamma shape is defined as

        .. math::

            f(x \\mid \\gamma, \\beta, \\mu) = (x - \\mu)^{\\gamma - 1} \\exp{\\left(-\\frac{x - \\mu}{\\beta}\\right)} / Z

        with the normalization over [0, inf] of

        .. math::

            Z = \\Gamma(\\gamma) \\beta^{\\gamma}

        The normalization changes for different normalization ranges and `Z=1` for the unnormalized shape.

        Args:
            gamma: Shape parameter of the gamma distribution
            beta: Scale parameter of the gamma distribution
            mu: Shift of the distribution
            obs: |@doc:pdf.init.obs| Observables of the
               model. This will be used as the default space of the PDF and,
               if not given explicitly, as the normalization range.

               The default space is used for example in the sample method: if no
               sampling limits are given, the default space is used.

               If the observables are binned and the model is unbinned, the
               model will be a binned model, by wrapping the model in a
               :py:class:`~zfit.pdf.BinnedFromUnbinnedPDF`, equivalent to
               calling :py:meth:`~zfit.pdf.BasePDF.to_binned`.

               If the observables are binned and the model is unbinned, the
               model will be a binned model, by wrapping the model in a
               :py:class:`~zfit.pdf.BinnedFromUnbinnedPDF`, equivalent to
               calling :py:meth:`~zfit.pdf.BasePDF.to_binned`.

               The observables are not equal to the domain as it does not restrict or
               truncate the model outside this range. |@docend:pdf.init.obs|
            extended: |@doc:pdf.init.extended| The overall yield of the PDF.
               If this is parameter-like, it will be used as the yield,
               the expected number of events, and the PDF will be extended.
               An extended PDF has additional functionality, such as the
               ``ext_*`` methods and the ``counts`` (for binned PDFs). |@docend:pdf.init.extended|
            norm: |@doc:pdf.init.norm| Normalization of the PDF.
               By default, this is the same as the default space of the PDF. |@docend:pdf.init.norm|
            name: |@doc:pdf.init.name| Name of the PDF.
               Maybe has implications on the serialization and deserialization of the PDF.
               For a human-readable name, use the label. |@docend:pdf.init.name|
            label: |@doc:pdf.init.label| Human-readable name
               or label of
               the PDF for a better description, to be used with plots etc.
               Has no programmatical functional purpose as identification. |@docend:pdf.init.label|
        """
        gamma, beta, mu = self._check_input_params_tfp(gamma, beta, mu)
        params = {"gamma": gamma, "beta": beta, "mu": mu}

        def dist_params():
            return {"concentration": gamma.value(), "rate": 1 / beta.value(), "loc": mu.value()}

        def distribution(concentration, rate, loc, name):
            return tfd.TransformedDistribution(
                distribution=tfp.distributions.Gamma(concentration, rate),
                bijector=tfp.bijectors.Shift(loc),
                name=name,
            )

        super().__init__(
            distribution=distribution,
            dist_params=dist_params,
            obs=obs,
            params=params,
            name=name,
            extended=extended,
            norm=norm,
            label=label,
        )


class GammaPDFRepr(BasePDFRepr):
    _implementation = Gamma
    hs3_type: Literal["Gamma"] = Field("Gamma", alias="type")
    x: SpaceRepr
    gamma: Serializer.types.ParamTypeDiscriminated
    beta: Serializer.types.ParamTypeDiscriminated
    mu: Serializer.types.ParamTypeDiscriminated


class JohnsonSU(WrapDistribution, SerializableMixin):
    _N_OBS = 1

    def __init__(
        self,
        mu: ztyping.ParamTypeInput,
        lambd: ztyping.ParamTypeInput,
        gamma: ztyping.ParamTypeInput,
        delta: ztyping.ParamTypeInput,
        obs: ztyping.ObsTypeInput,
        *,
        extended: ExtendedInputType = None,
        norm: NormInputType = None,
        name: str = "JohnsonSU",
        label: str | None = None,
    ):
        """Johnson's SU distribution.

        The Johnson SU shape is parametrized here with `mu`, `lambd`, `gamma` and `delta`, following
        the same parametrization `as RooFit <https://root.cern.ch/doc/master/classRooJohnson.html>`_.
        The Johnson SU shape results from transforming a normally distributed variable `x` to

        .. math::

            z = \\gamma + \\delta \\sinh^{-1}\\left(\\frac{x - \\mu}{\\lambda}\\right)

        The resulting shape is then

        .. math::

            f(x \\mid \\mu, \\lambda, \\gamma, \\delta) = \\exp{\\left[-\\frac{1}{2} \\left(\\gamma + \\delta \\sinh^{-1}\\left(\\frac{x - \\mu}{\\lambda}\\right)\\right)^2\\right]} / Z

        with the normalization over [-inf, inf] of

        .. math::

            Z = \\lambda \\sqrt{2 \\pi} \\sqrt{1 + \\left(\\frac{x - \\mu}{\\lambda}\\right)^2} / \\delta

        The normalization changes for different normalization ranges and `Z=1` for the unnormalized shape.

        Args:
            mu: Mean of the distribution
            lambd: Scale of the distribution
            gamma: Skewness of the distribution
            delta: Tailweight of the distribution
            obs: |@doc:pdf.init.obs| Observables of the
               model. This will be used as the default space of the PDF and,
               if not given explicitly, as the normalization range.

               The default space is used for example in the sample method: if no
               sampling limits are given, the default space is used.

               If the observables are binned and the model is unbinned, the
               model will be a binned model, by wrapping the model in a
               :py:class:`~zfit.pdf.BinnedFromUnbinnedPDF`, equivalent to
               calling :py:meth:`~zfit.pdf.BasePDF.to_binned`.

               If the observables are binned and the model is unbinned, the
               model will be a binned model, by wrapping the model in a
               :py:class:`~zfit.pdf.BinnedFromUnbinnedPDF`, equivalent to
               calling :py:meth:`~zfit.pdf.BasePDF.to_binned`.

               The observables are not equal to the domain as it does not restrict or
               truncate the model outside this range. |@docend:pdf.init.obs|
            extended: |@doc:pdf.init.extended| The overall yield of the PDF.
               If this is parameter-like, it will be used as the yield,
               the expected number of events, and the PDF will be extended.
               An extended PDF has additional functionality, such as the
               ``ext_*`` methods and the ``counts`` (for binned PDFs). |@docend:pdf.init.extended|
            norm: |@doc:pdf.init.norm| Normalization of the PDF.
               By default, this is the same as the default space of the PDF. |@docend:pdf.init.norm|
            name: |@doc:pdf.init.name| Name of the PDF.
               Maybe has implications on the serialization and deserialization of the PDF.
               For a human-readable name, use the label. |@docend:pdf.init.name|
            label: |@doc:pdf.init.label| Human-readable name
               or label of
               the PDF for a better description, to be used with plots etc.
               Has no programmatical functional purpose as identification. |@docend:pdf.init.label|
        """

        mu, lambd, gamma, delta = self._check_input_params_tfp(mu, lambd, gamma, delta)
        params = {"mu": mu, "lambd": lambd, "gamma": gamma, "delta": delta}

        def dist_params():
            return {"skewness": gamma.value(), "tailweight": delta.value(), "loc": mu.value(), "scale": lambd.value()}

        distribution = tfp.distributions.JohnsonSU
        super().__init__(
            distribution=distribution,
            dist_params=dist_params,
            obs=obs,
            params=params,
            name=name,
            extended=extended,
            norm=norm,
            label=label,
        )


class JohnsonSUPDFRepr(BasePDFRepr):
    _implementation = JohnsonSU
    hs3_type: Literal["JohnsonSU"] = Field("JohnsonSU", alias="type")
    x: SpaceRepr
    mu: Serializer.types.ParamTypeDiscriminated
    lambd: Serializer.types.ParamTypeDiscriminated
    gamma: Serializer.types.ParamTypeDiscriminated
    delta: Serializer.types.ParamTypeDiscriminated


class GeneralizedGauss(WrapDistribution, SerializableMixin):
    _N_OBS = 1

    def __init__(
        self,
        mu: ztyping.ParamTypeInput,
        sigma: ztyping.ParamTypeInput,
        beta: ztyping.ParamTypeInput,
        obs: ztyping.ObsTypeInput,
        *,
        extended: ExtendedInputType = None,
        norm: NormInputType = None,
        name: str = "GeneralizedGauss",
        label=None,
    ):
        """Generalized Gaussian distribution with a mean (mu), a standard deviation (sigma), and a shape parameter (beta).

        The gaussian shape is defined as

        .. math::
            f(x \\mid \\mu, \\sigma, \\beta) = e^{ -\\left(\\frac{|x - \\mu|}{\\sigma}\\right)^\\beta }

        with the normalization over [-inf, inf] of

        .. math::
            \\frac{1}{2 * \\beta * \\Gamma(1 + \\frac{1}{\\beta})}

        The normalization changes for different normalization ranges

        Args:
            mu: Mean of the gaussian dist
            sigma: Standard deviation or spread of the gaussian
            beta: Shape parameter of the generalized gaussian, which determines the shape of the tails
            obs: |@doc:pdf.init.obs| Observables of the
               model. This will be used as the default space of the PDF and,
               if not given explicitly, as the normalization range.

               The default space is used for example in the sample method: if no
               sampling limits are given, the default space is used.

               If the observables are binned and the model is unbinned, the
               model will be a binned model, by wrapping the model in a
               :py:class:`~zfit.pdf.BinnedFromUnbinnedPDF`, equivalent to
               calling :py:meth:`~zfit.pdf.BasePDF.to_binned`.

               If the observables are binned and the model is unbinned, the
               model will be a binned model, by wrapping the model in a
               :py:class:`~zfit.pdf.BinnedFromUnbinnedPDF`, equivalent to
               calling :py:meth:`~zfit.pdf.BasePDF.to_binned`.

               The observables are not equal to the domain as it does not restrict or
               truncate the model outside this range. |@docend:pdf.init.obs|
            extended: |@doc:pdf.init.extended| The overall yield of the PDF.
               If this is parameter-like, it will be used as the yield,
               the expected number of events, and the PDF will be extended.
               An extended PDF has additional functionality, such as the
               ``ext_*`` methods and the ``counts`` (for binned PDFs). |@docend:pdf.init.extended|
            norm: |@doc:pdf.init.norm| Normalization of the PDF.
               By default, this is the same as the default space of the PDF. |@docend:pdf.init.norm|
            name: |@doc:pdf.init.name| Name of the PDF.
               Maybe has implications on the serialization and deserialization of the PDF.
               For a human-readable name, use the label. |@docend:pdf.init.name|
            label: |@doc:pdf.init.label| Human-readable name
               or label of
               the PDF for a better description, to be used with plots etc.
               Has no programmatical functional purpose as identification. |@docend:pdf.init.label|
        """
        mu, sigma, beta = self._check_input_params_tfp(mu, sigma, beta)
        params = {"mu": mu, "sigma": sigma, "beta": beta}

        def dist_params():
            return {"loc": mu.value(), "scale": sigma.value(), "power": beta.value()}

        distribution = tfp.distributions.GeneralizedNormal
        super().__init__(
            distribution=distribution,
            dist_params=dist_params,
            obs=obs,
            params=params,
            name=name,
            extended=extended,
            norm=norm,
            label=label,
        )


class GeneralizedGaussPDFRepr(BasePDFRepr):
    _implementation = GeneralizedGauss
    hs3_type: Literal["GeneralizedGauss"] = Field("GeneralizedGauss", alias="type")
    x: SpaceRepr
    mu: Serializer.types.ParamTypeDiscriminated
    sigma: Serializer.types.ParamTypeDiscriminated
    beta: Serializer.types.ParamTypeDiscriminated



#  Copyright (c) 2024 zfit
import numpy as np
import pytest

import zfit
import zfit.z.numpy as znp
from zfit import supports


def createpdf1(obs, params):
    params = params.copy()
    yieldp = params.pop('yield')
    class SomeGradsPDF(zfit.pdf.BasePDF):

        def __init__(self, obs, a, b, c, d, e, f, name="SomeGradsPDF", **kwargs):
            params = {'a': a, 'b': b, 'c': c, 'd': d, 'e':e, 'f':f}
            super().__init__(obs=obs, params=params, name=name, autograd_params=['a', 'b', 'yield', 'e'], **kwargs)

        @supports(norm=False)
        def _pdf(self, x, norm, *, params=None):
            a = params['a']
            b = params['b']
            c = params['c']
            d = params['d']
            e = params['e']
            f = params['f']
            x = x[0]
            return (znp.exp(-a * x) + znp.exp(-b * x) + c * x + d * x + e * x + f * x) ** 2
    return SomeGradsPDF(obs=obs, **params, extended=yieldp)

def createpdfbinned1(obs, params):
    obs = obs.with_binning(43)
    pdf = createpdf1(obs, params)
    return pdf

def createsum1(obs, params):
    pdf1 = createpdf1(obs, params)
    pdf2 = zfit.pdf.Gauss(obs=obs, mu=params['a'], sigma=params['b'], extended=params['yield'])
    return zfit.pdf.SumPDF([pdf1, pdf2], fracs=params['c'], extended=params['yield'])

def createbinnedsum1(obs, params):
    obs = obs.with_binning(12)
    pdf = createsum1(obs, params)
    return pdf

def createbinnedsum2(obs, params):
    obs = obs.with_binning(12)
    pdf1 = createpdf1(obs, params)
    pdf2 = zfit.pdf.Gauss(obs=obs, mu=params['a'], sigma=params['b'], extended=params['yield'])
    return zfit.pdf.BinnedSumPDF([pdf1, pdf2], fracs=params['c'], extended=params['yield'])

def createnll1(obs,params):
    pdf = createpdf1(obs,params)
    data = pdf.sample(1000)
    nll = zfit.loss.ExtendedUnbinnedNLL(model=pdf, data=data)
    return nll

def createbinnednll1(obs,params):
    obs = obs.with_binning(12)
    pdf = createpdf1(obs,params)
    data = pdf.sample(1000)
    nll = zfit.loss.ExtendedBinnedNLL(model=pdf, data=data)
    return nll


@pytest.mark.parametrize('paramdepfactory', [
    createpdf1,
    createpdfbinned1,
    createsum1,
    createbinnedsum1,
    createbinnedsum2,
    createnll1,
    createbinnednll1

])
def test_nograd_params(paramdepfactory):


    parama = zfit.Parameter("param_a", 1.)
    paramb = zfit.Parameter("param_b", 2., floating=False)
    yieldp = zfit.Parameter("yieldp", 3.)
    parameindep = zfit.Parameter("param_eindep", 6., floating=True)
    pgrad = [parama, paramb,
             yieldp, parameindep]
    parame = zfit.ComposedParameter("param_e", lambda params: params[0] + params[1], params=[pgrad[0], pgrad[-1]])
    pgrad.append(parame)
    paramc = zfit.Parameter("param_c", 0.4, floating=True)
    paramd = zfit.Parameter("param_d", 5., floating=False)
    paramfindep = zfit.Parameter("param_findep", 3., floating=True)
    pnograd = [paramc, paramd,paramfindep]
    paramf = zfit.ComposedParameter("param_f", lambda params: params[0] + params[1],
                                       params=[pnograd[0], pnograd[-1]])
    pnograd.append(paramf)
    obs = zfit.Space('obs1', -1, 5)
    pdf = paramdepfactory(obs, {'a': parama, 'b': paramb, 'c': paramc, 'd': paramd, 'e': parame, 'f': paramf, 'yield': yieldp})
    # yieldp = pdf.get_yield()


    pgrad = set(pgrad)
    pnograd = set(pnograd)
    params = pdf.get_params(is_yield=None, autograd=None, floating=None, extract_independent=None)
    compset = set()  # add a frac composed parameter if autogenerated, supports gradients
    for p in params:
        if p not in pgrad and p not in pnograd:
            if p.name == 'Composed_autoparam_0':  # add the fraction parameter that is autocombosed
                compset = {p}
                pgrad |= compset
    assert pgrad == set(pdf.get_params(autograd=True, floating=None, extract_independent=None))
    assert pnograd == set(pdf.get_params(autograd=False, floating=None, extract_independent=None))
    assert pgrad | pnograd == params
    assert pgrad & pnograd == set()
    assert {yieldp} == pdf.get_params(is_yield=True)
    assert {yieldp} == pdf.get_params(is_yield=True, autograd=True)
    assert set() == pdf.get_params(is_yield=True, autograd=False)
    assert {parama, paramb, yieldp, parameindep, parame} | compset == set(pdf.get_params(is_yield=None, autograd=True, floating=None, extract_independent=None))
    assert {parama, paramb, yieldp, parameindep} == set(pdf.get_params(is_yield=None, autograd=True, floating=None, extract_independent=True))
    assert {parama, yieldp, parameindep} == set(pdf.get_params(is_yield=None, autograd=True, floating=True, extract_independent=True))
    assert {parama, paramb, yieldp, parame} | compset == set(pdf.get_params(is_yield=None, autograd=True, floating=None, extract_independent=False))

    assert {paramc, paramd, paramfindep} == set(pdf.get_params(is_yield=None, autograd=False, floating=None, extract_independent=True))
    assert {paramc, paramd, paramf} == set(pdf.get_params(is_yield=None, autograd=False, floating=None, extract_independent=False))
    assert {paramc, paramd, paramfindep, paramf} == set(pdf.get_params(is_yield=None, autograd=False, floating=None, extract_independent=None))
    assert set() == set(pdf.get_params(is_yield=True, autograd=False, floating=None, extract_independent=None))
    assert {paramc, paramfindep} == set(pdf.get_params(is_yield=None, autograd=False, floating=True, extract_independent=True))
    assert {paramd} == set(pdf.get_params(is_yield=None, autograd=False, floating=False, extract_independent=True))

@pytest.mark.parametrize('lossfactory', [createbinnednll1, createnll1])
def test_autograd_nogradparams_raises(lossfactory):
    parama = zfit.Parameter("param_a", 1.)
    paramb = zfit.Parameter("param_b", 2., floating=False)
    yieldp = zfit.Parameter("yieldp", 1000.)
    parameindep = zfit.Parameter("param_eindep", 6., floating=True)
    pgrad = [parama, paramb,
             yieldp, parameindep]
    parame = zfit.ComposedParameter("param_e", lambda params: params[0] * params[1], params=[pgrad[0], parameindep])
    paramc = zfit.Parameter("param_c", 0.4, floating=True)
    paramd = zfit.Parameter("param_d", 5., floating=False)
    paramfindep = zfit.Parameter("param_findep", 3., floating=True)
    pnograd = [paramc, paramd,paramfindep]
    paramf = zfit.ComposedParameter("param_f", lambda params: params[0] + params[1],
                                       params=[pnograd[0], paramfindep])
    obs = zfit.Space('obs1', -1, 5)
    loss = lossfactory(obs, {'a': parama, 'b': paramb, 'c': paramc, 'd': paramd, 'e': parame, 'f': paramf, 'yield': yieldp})
    value = loss.value()
    assert not np.isnan(value), "Test is not valid if value is nan"
    with pytest.raises(zfit.exception.AutogradNotSupported):
        loss.gradient()
    with pytest.raises(zfit.exception.AutogradNotSupported):
        loss.gradient(params=pnograd)
    with pytest.raises(zfit.exception.AutogradNotSupported):
        loss.gradient(params=pnograd[0])
    _ = loss.gradient(params=pnograd, numgrad=True)
    _ = loss.gradient(numgrad=True)
    gradient1 = loss.gradient(params=pgrad, numgrad=True)
    gradient2 = loss.gradient(params=pgrad, numgrad=False)
    assert np.allclose(gradient1, gradient2, rtol=1e-5, atol=1e-5)

    hessian1 = loss.hessian(params=pgrad, numgrad=True)
    hessian2 = loss.hessian(params=pgrad, numgrad=False)
    np.testing.assert_allclose(hessian1, hessian2, atol=1e-4, rtol=1e10)

    with pytest.raises(zfit.exception.AutogradNotSupported):
        _ = loss.hessian(params=pnograd, numgrad=False)
    hessian3 = loss.hessian(params=pnograd, numgrad=True)
    assert not np.any(np.isnan(hessian3)), f"NaN in hessian: {hessian3}"
    with pytest.raises(zfit.exception.AutogradNotSupported):
        _ = loss.hessian(params=loss.get_params(), numgrad=False)



#  Copyright (c) 2024 zfit
import numpy as np
import pytest
from scipy.stats import crystalball

import zfit
import zfit.z.numpy as znp
from zfit.core.testing import tester
from zfit.models.physics import CrystalBall, DoubleCB, GeneralizedCB

mu = -0.3
sigma = 1.1

sigmal = 1.1
sigmar = 1.1

alphal = 0.8
nl = 2.0

alphar = 1.4
nr = 4.0

bounds = (-6, 4)
lbounds = (bounds[0], mu)
rbounds = (mu, bounds[1])


def _cb_params_factory(name_add=""):
    mu_ = zfit.Parameter(f"mu_cb{name_add}", mu)
    sigma_ = zfit.Parameter(f"sigma_cb{name_add}", sigma)
    alphal_ = zfit.Parameter(f"alphal_cb{name_add}", alphal)
    nl_ = zfit.Parameter(f"nl_cb{name_add}", nl)
    return {"mu": mu_, "sigma": sigma_, "alpha": alphal_, "n": nl_}


tester.register_pdf(pdf_class=CrystalBall, params_factories=_cb_params_factory)


def sample_testing(pdf):
    sample = pdf.sample(n=1000, limits=(-0.5, 1.5))
    assert not any(np.isnan(sample.value()))


def eval_testing(pdf, x):
    probs = pdf.pdf(x)
    assert probs.shape.rank == 1
    assert probs.shape[0] == x.shape[0]
    probs = znp.asarray(probs)
    assert not np.any(np.isnan(probs))
    return probs


def test_cb_integral():
    obs = zfit.Space("x", limits=bounds)

    mu_ = zfit.Parameter("mu", mu)
    sigma_ = zfit.Parameter("sigma", sigma)
    alphal_ = zfit.Parameter("alphal", alphal)
    nl_ = zfit.Parameter("nl", nl)

    cbl = CrystalBall(obs=obs, mu=mu_, sigma=sigma_, alpha=alphal_, n=nl_)
    int_limits = (-1, 3)
    integral_numeric = cbl.numeric_integrate(limits=int_limits, norm=False)

    integral = cbl.analytic_integrate(limits=int_limits, norm=False)

    assert pytest.approx(integral_numeric, 1e-5) == integral

    rnd_limits = sorted(list(np.random.uniform(*bounds, 13)) + list(bounds))
    integrals = [
        cbl.integrate((low, up), norm=False)
        for low, up in zip(rnd_limits[:-1], rnd_limits[1:])
    ]

    integral = np.sum(integrals)
    integral_full = (cbl.integrate(bounds, norm=False))
    assert pytest.approx(integral_full) == (integral)


@pytest.mark.parametrize("doublecb", ["DoubleCB", "GeneralizedCB"])
def test_cb_dcb(doublecb):
    obs = zfit.Space("x", limits=bounds)

    mu_ = zfit.Parameter("mu", mu)
    sigma_ = zfit.Parameter("sigma", sigma)
    sigmal_ = zfit.Parameter("sigmal", sigmal)
    alphal_ = zfit.Parameter("alphal", alphal)
    nl_ = zfit.Parameter("nl", nl)
    sigmar_ = zfit.Parameter("sigmar", sigmar)
    alphar_ = zfit.Parameter("alphar", alphar)
    nr_ = zfit.Parameter("nr", nr)

    cbl = CrystalBall(obs=obs, mu=mu_, sigma=sigmal_, alpha=alphal_, n=nl_)
    cbr = CrystalBall(obs=obs, mu=mu_, sigma=sigmar_, alpha=-alphar_, n=nr_)
    if doublecb == "DoubleCB":
        dcb = DoubleCB(
            obs=obs,
            mu=mu_,
            sigma=sigma_,
            alphal=alphal_,
            nl=nl_,
            alphar=alphar_,
            nr=nr_,
        )
    else:
        dcb = GeneralizedCB(
            obs=obs,
            mu=mu_,
            sigmal=sigmal_,
            alphal=alphal_,
            nl=nl_,
            sigmar=sigmar_,
            alphar=alphar_,
            nr=nr_,
        )

    sample_testing(cbl)
    sample_testing(cbr)
    sample_testing(dcb)

    x = np.random.normal(mu, sigma, size=10_000)

    probsl = eval_testing(cbl, x)
    probsr = eval_testing(cbr, x)

    assert not any(np.isnan(probsl))
    assert not any(np.isnan(probsr))

    probsl_scipy = crystalball.pdf(x, beta=alphal, m=nl, loc=mu, scale=sigma)
    probsr_scipy = crystalball.pdf(-x + 2 * mu, beta=alphar, m=nr, loc=mu, scale=sigma)

    # We take the ration as the normalization is not fixed
    ratio_l = probsl_scipy / probsl
    ratio_r = probsr_scipy / probsr

    ones = np.ones_like(ratio_l)
    np.testing.assert_allclose(
        ratio_l,
        ones * ratio_l[0],
        rtol=5e-7,
        atol=5e-7,
    )  # shape should be same
    np.testing.assert_allclose(ratio_r, ones * ratio_r[0], rtol=5e-7, atol=5e-7)

    kwargs = dict(limits=(-5.0, mu), norm=lbounds)
    intl = cbl.integrate(**kwargs) - dcb.integrate(**kwargs)
    assert pytest.approx(intl, abs=1e-3) == 0.0
    intl = cbr.integrate(**kwargs) - dcb.integrate(**kwargs)
    assert pytest.approx(intl, abs=1e-3) != 0

    # TODO: update test to fixed DCB integral
    kwargs = dict(limits=(mu, 2.0), norm=rbounds)
    dcb_integr1 = dcb.integrate(**kwargs)
    intr = cbr.integrate(**kwargs) - dcb_integr1
    assert pytest.approx(intr, abs=1e-3) == 0.0
    intr = cbl.integrate(**kwargs) - dcb.integrate(**kwargs)
    assert pytest.approx(intr, abs=1e-3) != 0.0

    xl = x[x <= mu]
    xr = x[x > mu]

    probs_dcb_l = eval_testing(dcb, xl)
    probs_dcb_r = eval_testing(dcb, xr)

    probsl_scipy = crystalball.pdf(xl, beta=alphal, m=nl, loc=mu, scale=sigma)
    probsr_scipy = crystalball.pdf(-xr + 2 * mu, beta=alphar, m=nr, loc=mu, scale=sigma)

    ratio_l = probsl_scipy / probs_dcb_l
    ratio_r = probsr_scipy / probs_dcb_r

    np.testing.assert_allclose(ratio_l, ratio_l[0], rtol=3e-7)
    np.testing.assert_allclose(ratio_r, ratio_r[0], rtol=3e-7)

    rnd_limits = sorted(list(np.random.uniform(*bounds, 130)) + list(bounds))
    integrals = []
    for low, up in zip(rnd_limits[:-1], rnd_limits[1:]):
        integrals.append(dcb.integrate((low, up), norm=False))

    integral = np.sum(integrals)
    integral_full = znp.asarray(dcb.integrate((bounds[0], up), norm=False))
    assert pytest.approx(float(integral_full)) == float(integral)



from __future__ import annotations

from .core.interfaces import (
    ZfitBinnedData,
    ZfitBinnedPDF,
    ZfitBinning,
    ZfitData,
    ZfitIndependentParameter,
    ZfitLoss,
    ZfitModel,
    ZfitParameter,
    ZfitPDF,
    ZfitSpace,
    ZfitUnbinnedData,
)

__all__ = [
    "ZfitData",
    "ZfitModel",
    "ZfitLoss",
    "ZfitParameter",
    "ZfitSpace",
    "ZfitUnbinnedData",
    "ZfitIndependentParameter",
    "ZfitBinning",
    "ZfitBinnedPDF",
    "ZfitBinnedData",
    "ZfitPDF",
]



#  Copyright (c) 2024 zfit
from __future__ import annotations

import tensorflow as tf

import zfit.z.numpy as znp

from ..settings import ztypes


def generate_1d_grid(data, num_grid_points, absolute_boundary=0.0, relative_boundary=0.05):
    absolute_boundary = znp.asarray(absolute_boundary, ztypes.float)
    relative_boundary = znp.asarray(relative_boundary, ztypes.float)
    num_grid_points = znp.asarray(num_grid_points, ztypes.int)
    minimum = znp.min(data)
    maximum = znp.max(data)
    space_width = maximum - minimum
    outside_borders = znp.maximum(relative_boundary * space_width, absolute_boundary)

    return tf.linspace(
        minimum - outside_borders, maximum + outside_borders, num=num_grid_points
    )  # requires tf, znp doesn't work
    # znp requires num_grid_points to be a python int and then can't do the division (as the others are floats)


def bin_1d(binning_method, data, grid, weights=None):
    if binning_method == "simple":
        return bin_1d_simple(data, grid, weights)
    elif binning_method == "linear":
        return bin_1d_linear(data, grid, weights)
    else:
        msg = f"Binning method '{binning_method}' not supported, only 'simple' or 'linear'."
        raise ValueError(msg)


def bin_1d_simple(data, grid, weights=None):
    if weights is None:
        bincount = znp.asarray(
            tf.histogram_fixed_width(
                data,
                [znp.min(grid), znp.max(grid)],
                znp.size(grid),
            ),
            ztypes.float,
        )
    else:
        bincount = _bin_1d_weighted(data, grid, weights, "simple")

    return bincount


def bin_1d_linear(data, grid, weights=None):
    return _bin_1d_weighted(data, grid, weights)


def _bin_1d_weighted(data, grid, weights, method="linear"):
    if weights is None:
        weights = znp.ones_like(data, ztypes.float)

    weights = weights / znp.sum(weights)

    grid_size = znp.size(grid)
    grid_min = znp.min(grid)
    grid_max = znp.max(grid)
    num_intervals = grid_size - tf.constant(1)
    dx = (grid_max - grid_min) / znp.asarray(num_intervals, ztypes.float)

    transformed_data = (data - grid_min) / dx

    # Compute the integral and fractional part of the data
    # The integral part is used for lookups, the fractional part is used
    # to weight the data
    integral = znp.floor(transformed_data)
    fractional = transformed_data - integral

    if method == "simple":
        fractional = znp.asarray(fractional > 0.5, fractional.dtype) * fractional

    # Compute the weights for left and right side of the linear binning routine
    frac_weights = fractional * weights
    neg_frac_weights = weights - frac_weights

    # tf.math.bincount only works with tf.int32
    bincount_left = tf.roll(
        tf.concat(
            tf.math.bincount(
                znp.asarray(integral, tf.int32),
                weights=frac_weights,
                minlength=grid_size,
                maxlength=grid_size,
            ),
            tf.constant(0),
        ),
        shift=1,
        axis=0,
    )
    bincount_right = tf.math.bincount(
        znp.asarray(integral, tf.int32),
        weights=neg_frac_weights,
        minlength=grid_size,
        maxlength=grid_size,
    )

    return znp.asarray(bincount_left + bincount_right, ztypes.float)
